package net.sf.jpatch;

import java.awt.*;
import java.awt.image.*;
import java.io.*;
import java.util.*;
import javax.swing.*;
import javax.vecmath.*;

import net.sf.jpatch.boundary.*;
import net.sf.jpatch.entity.*;

public class Render extends JPanel {
	
	private static IZBufferRenderer renderer;
	private static int iSupersampling = 2;
	private static int iWidth = 700;
	private static int iHeight = 700;
	private static Image im;
	
	public Render(String filename) {
		renderer = new IZBufferRenderer(iWidth * iSupersampling,iHeight * iSupersampling,0);
		MaterialProperties mp = new MaterialProperties();
		mp.specular = 1f;
		mp.roughness = 0.01f;
		
		
		Matrix4d matrix = new Matrix4d();
		matrix.setIdentity();
		
		
		
		translate(matrix,0,0,-400);
		rotate(matrix,-10,-30,0);
		translate(matrix,0,120,0);
		
		//translate(matrix,0,0,-2000);
		
		
		renderer.setCamera(matrix, iWidth * iSupersampling / 35 * 20);	// 50mm lense (on 35mm film)
		
		renderer.setAmbientLightColor(new Color3f(1,1,1));
		renderer.setNumberOfLights(3);
		
		/* key */
		renderer.addLightSource(new Vector3d(-1,1,-1), new Color3f(0.6f,0.6f,0.6f), true);
		
		/* fill */
		renderer.addLightSource(new Vector3d(1,0,-1), new Color3f(0.3f,0.3f,0.3f), false);
		
		/* back */
		renderer.addLightSource(new Vector3d(0,1,1), new Color3f(1,1,1), true);
		
		renderFile(filename, renderer);
	}
	
	public void rotate(Matrix4d m, double x, double y, double z) {
		AxisAngle4d a;
		Matrix4d t = new Matrix4d();
		a = new AxisAngle4d(1,0,0,x / 180 * Math.PI);
		t.setIdentity();
		t.setRotation(a);
		m.mul(t);
		a = new AxisAngle4d(0,1,0,y / 180 * Math.PI);
		t.setIdentity();
		t.setRotation(a);
		m.mul(t);
		a = new AxisAngle4d(0,0,1,z / 180 * Math.PI);
		t.setIdentity();
		t.setRotation(a);
		m.mul(t);
	}
	
	public void translate(Matrix4d m, double x, double y, double z) {
		Matrix4d t = new Matrix4d();
		t.setIdentity();
		t.setTranslation(new Vector3d(-x,-y,-z));
		m.mul(t);
	}
		
	public void renderFile(String filename, IZBufferRenderer renderer) {
		BufferedReader brFile;
		String strLine;
		String[] astrPart;
		double x,y,z;
		ArrayList listPoint = new ArrayList();
		ArrayList listNormal = new ArrayList();
		MaterialProperties mp = new MaterialProperties();
		try {
			brFile = new BufferedReader(new FileReader(filename));
			while ((strLine = brFile.readLine()) != null) {
				
				/* read point position and normal */
				if (strLine.startsWith("p")) {
					astrPart = strLine.split("\\s");
					
					x = (new Double(astrPart[1])).doubleValue();
					y = (new Double(astrPart[2])).doubleValue();
					z = (new Double(astrPart[3])).doubleValue();
					Point3d point = new Point3d(x,y,z);
					
					x = (new Double(astrPart[4])).doubleValue();
					y = (new Double(astrPart[5])).doubleValue();
					z = (new Double(astrPart[6])).doubleValue();
					Vector3d normal = new Vector3d(x,y,z);
					
					renderer.transform(point,normal);
					listPoint.add(point);
					listNormal.add(normal);
				} else
				
				/* read material properties */
				if (strLine.startsWith("m")) {
					astrPart = strLine.split("\\s");
					mp.red = (new Float(astrPart[1])).floatValue();
					mp.green = (new Float(astrPart[2])).floatValue();
					mp.blue = (new Float(astrPart[3])).floatValue();
					mp.ambient = (new Float(astrPart[4])).floatValue();
					mp.diffuse = (new Float(astrPart[5])).floatValue();
					mp.brilliance = (new Float(astrPart[6])).floatValue();
					mp.specular = (new Float(astrPart[7])).floatValue();
					mp.roughness = (new Float(astrPart[8])).floatValue();
				} else
				
				/* read and render face */
				if (strLine.startsWith("f")) {
					astrPart = strLine.split("\\s");
					int a = (new Integer(astrPart[1])).intValue();
					int b = (new Integer(astrPart[2])).intValue();
					int c = (new Integer(astrPart[3])).intValue();
					renderer.draw3DTrianglePhong(
						(Point3d) listPoint.get(a),
						(Point3d) listPoint.get(b),
						(Point3d) listPoint.get(c),
						(Vector3d) listNormal.get(a),
						(Vector3d) listNormal.get(b),
						(Vector3d) listNormal.get(c),
						mp
					);
				}
			}
		} catch (IOException e) {
			System.err.println(e);
		}
	}
						
					
	public void paint(Graphics g) {
		g.drawImage(im,0,0,null);
	}
	
	public static void main(String[] args) {
		Render r = new Render(args[0]);
		Image i = renderer.getImage();
		im = i.getScaledInstance(iWidth,iHeight, Image.SCALE_AREA_AVERAGING);
		JFrame frame = new JFrame();
		frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		r.setPreferredSize(new Dimension(iWidth,iHeight));
		frame.getContentPane().add(r);
		frame.pack();
		frame.show();
	}
}

