package net.sf.jpatch;

import java.awt.*;
import java.awt.image.*;
import java.io.*;
import java.util.*;
import javax.swing.*;
import javax.vecmath.*;
import javax.imageio.*;

import net.sf.jpatch.boundary.*;
import net.sf.jpatch.entity.*;
import net.sf.jpatch.auxilary.*;

public class Render2 extends JPanel implements Shader {
	
	private int iWidth;
	private int iHeight;
	private int iSupersample;
	private int iW;
	private int iH;
	
	private int iCenterX;
	private int iCenterY;
	private BufferedImage image;
	private int[] aiFrameBuffer;
	
	private Matrix4d m4View;
	private double dFocalLength;
	
	private Vector3d[] av3LightDirection;
	private Color3f[] ac3LightColor;
	private boolean[] abLightSpecular;
	private Color3f c3AmbientLight;
	private int iLights;
	private int iLightNum;
	
	private ArrayList listTriangles = new ArrayList();
	
	public Render2(int width, int height, int supersample) {
		
		iWidth = width;
		iHeight = height;
		iSupersample = supersample;
		
		iW = width * supersample;
		iH = height * supersample;
		
		iCenterX = iW / 2;
		iCenterY = iH / 2;

		image = new BufferedImage(iWidth, iHeight, BufferedImage.TYPE_INT_RGB);
		aiFrameBuffer = ((DataBufferInt) image.getRaster().getDataBuffer()).getData();
		
		m4View = new Matrix4d();
		m4View.setIdentity();
		
		dFocalLength = (double) (iW) / 35 * 50;
		
		translate(m4View,0,0,-500/3/3);
		rotate(m4View,-10,-20,0);
		translate(m4View,0,155,0);
		
		//translate(m4View,0,0,-50);
		
		
		//renderer.setCamera(matrix, iWidth * iSupersampling / 35 * 50);	// 50mm lense (on 35mm film)
		
		c3AmbientLight = new Color3f(1,1,1);
		iLights = 3;
		av3LightDirection = new Vector3d[iLights];
		ac3LightColor = new Color3f[iLights];
		abLightSpecular = new boolean[iLights];

		/* key */
		addLightSource(new Vector3d(-1,1,-1), new Color3f(0.6f,0.6f,0.6f), true);
		
		/* fill */
		addLightSource(new Vector3d(1,0,-1), new Color3f(0.3f,0.3f,0.3f), false);
		
		/* back */
		addLightSource(new Vector3d(0,1,1), new Color3f(1,1,1), true);
	}
	
	public void addLightSource(Vector3d direction, Color3f color, boolean specular) {
		av3LightDirection[iLightNum] = direction;
		m4View.transform(av3LightDirection[iLightNum]);
		av3LightDirection[iLightNum].normalize();
		ac3LightColor[iLightNum] = color;
		abLightSpecular[iLightNum] = specular;
		iLightNum++;
	}
	
	public void rotate(Matrix4d m, double x, double y, double z) {
		AxisAngle4d a;
		Matrix4d t = new Matrix4d();
		a = new AxisAngle4d(1,0,0,x / 180 * Math.PI);
		t.setIdentity();
		t.setRotation(a);
		m.mul(t);
		a = new AxisAngle4d(0,1,0,y / 180 * Math.PI);
		t.setIdentity();
		t.setRotation(a);
		m.mul(t);
		a = new AxisAngle4d(0,0,1,z / 180 * Math.PI);
		t.setIdentity();
		t.setRotation(a);
		m.mul(t);
	}
	
	public void translate(Matrix4d m, double x, double y, double z) {
		Matrix4d t = new Matrix4d();
		t.setIdentity();
		t.setTranslation(new Vector3d(-x,-y,-z));
		m.mul(t);
	}
		
	public void initFromFile(String filename) {
		BufferedReader brFile;
		String strLine;
		String[] astrPart;
		double x,y,z;
		ArrayList listPoint = new ArrayList();
		ArrayList listNormal = new ArrayList();
		MaterialProperties mp = new MaterialProperties();
		try {
			brFile = new BufferedReader(new FileReader(filename));
			while ((strLine = brFile.readLine()) != null) {
				
				/* read point position and normal */
				if (strLine.startsWith("p")) {
					astrPart = strLine.split("\\s");
					
					x = (new Double(astrPart[1])).doubleValue();
					y = (new Double(astrPart[2])).doubleValue();
					z = (new Double(astrPart[3])).doubleValue();
					Point3d point = new Point3d(x,y,z);
					
					x = (new Double(astrPart[4])).doubleValue();
					y = (new Double(astrPart[5])).doubleValue();
					z = (new Double(astrPart[6])).doubleValue();
					Vector3d normal = new Vector3d(x,y,z);
					
					//renderer.transform(point,normal);
					m4View.transform(point);
					m4View.transform(normal);
					
					/*
					* flip backface normals
					*/
					//Vector3d D = new Vector3d(point);
					//D.normalize();
					//if (D.dot(normal) >= 0) {
					//	normal.x = -normal.x;
					//	normal.y = -normal.y;
					//	normal.z = -normal.z;
					//}
		
					listPoint.add(point);
					listNormal.add(normal);
				} else
				
				/* read material properties */
				if (strLine.startsWith("m")) {
					mp = new MaterialProperties();
					astrPart = strLine.split("\\s");
					mp.red = (new Float(astrPart[1])).floatValue();
					mp.green = (new Float(astrPart[2])).floatValue();
					mp.blue = (new Float(astrPart[3])).floatValue();
					mp.ambient = (new Float(astrPart[4])).floatValue();
					mp.diffuse = (new Float(astrPart[5])).floatValue();
					mp.brilliance = (new Float(astrPart[6])).floatValue();
					mp.specular = (new Float(astrPart[7])).floatValue();
					mp.roughness = (new Float(astrPart[8])).floatValue();
				} else
				
				/* read face */
				if (strLine.startsWith("f")) {
					astrPart = strLine.split("\\s");
					int a = (new Integer(astrPart[1])).intValue();
					int b = (new Integer(astrPart[2])).intValue();
					int c = (new Integer(astrPart[3])).intValue();
					addTriangle(
						(Point3d) listPoint.get(a),
						(Point3d) listPoint.get(b),
						(Point3d) listPoint.get(c),
						(Vector3d) listNormal.get(a),
						(Vector3d) listNormal.get(b),
						(Vector3d) listNormal.get(c),
						mp
					);
				}
				
				/* read patch */
				if (strLine.startsWith("b")) {
					float f = (float) dFocalLength;
					astrPart = strLine.split("\\s");
					Point3f[] bezierControlPoints = new Point3f[16];
					int n = 1;
					float minx = Float.MAX_VALUE;
					float maxx = -Float.MAX_VALUE;
					float miny = Float.MAX_VALUE;
					float maxy = -Float.MAX_VALUE;
					for (int i = 0; i < 16; i++) {
						Point3f p = new Point3f((new Float(astrPart[n++])).floatValue(),(new Float(astrPart[n++])).floatValue(),(new Float(astrPart[n++])).floatValue());
						m4View.transform(p);
						bezierControlPoints[i] = new Point3f(p);
						if (p.z < 1) p.z = 1;
						p.x /= p.z / f;
						p.y /= p.z / f;
						if (p.x > maxx) maxx = p.x;
						if (p.y > maxy) maxy = p.y;
						if (p.x < minx) minx = p.x;
						if (p.y < miny) miny = p.y;
					}
					float dx = (maxx - minx) / iSupersample;
					float dy = (maxy - miny) / iSupersample;
					dx = (float) Math.sqrt(dx) + 1;
					dy = (float) Math.sqrt(dy) + 1;
					int subdiv = (dx > dy) ? (int) dx : (int) dy;
					if (subdiv > 8) subdiv = 8;
					if (subdiv < 3) subdiv = 3;
					System.out.println(dx + " " + dy + " " + subdiv);
					//subdiv = 4;
					Tesselator.tesselate(bezierControlPoints,subdiv);
					Point3d[][] p3P = Tesselator.getPointArray();
					Vector3d[][] v3N = Tesselator.getNormalArray();
					for (int u = 0; u < subdiv - 1; u++) {
						for (int v = 0; v < subdiv - 1; v++) {
							addTriangle(p3P[u][v], p3P[u + 1][v], p3P[u][v + 1],v3N[u][v], v3N[u + 1][v], v3N[u][v + 1], mp);
							addTriangle(p3P[u + 1][v], p3P[u + 1][v + 1], p3P[u][v + 1],v3N[u + 1][v], v3N[u + 1][v + 1], v3N[u][v + 1], mp);
						}
					}
				}
			}
		} catch (IOException e) {
			System.err.println(e);
		}
	}
	
	public void addTriangle(Point3d pa, Point3d pb, Point3d pc, Vector3d na, Vector3d nb, Vector3d nc, MaterialProperties mp) {
		//System.out.println("addTriangle()");
		Point3d A = new Point3d();
		Point3d B = new Point3d();
		Point3d C = new Point3d();
		Point3d D = new Point3d();
		Point3d pd = new Point3d();
		Vector3d nd = new Vector3d();
		A.x = pa.x / pa.z * dFocalLength + iCenterX;
		A.y = -pa.y / pa.z * dFocalLength + iCenterY;
		B.x = pb.x / pb.z * dFocalLength + iCenterX;
		B.y = -pb.y / pb.z * dFocalLength + iCenterY;
		C.x = pc.x / pc.z * dFocalLength + iCenterX;
		C.y = -pc.y / pc.z * dFocalLength + iCenterY;
		//System.out.println(A.x + " " + A.y + " " + A.z);
		//System.out.println(B.x + " " + B.y + " " + B.z);
		//System.out.println(C.x + " " + C.y + " " + C.z);
		//System.out.println(iWidth + " " + iHeight);
		if (A.x >= 0 && A.x < iW && A.y >= 0 && A.y < iH || B.x >= 0 && B.x < iW && B.y >= 0 && B.y < iH || C.x >= 0 && C.x < iW && C.y >= 0 && C.y < iH) {
			//System.out.println("*");
			if (A.y < B.y && A.y < C.y) {
				//System.out.println("A < ..");
				if (B.y < C.y) {
					//System.out.println("..B < C");
					double f = (B.y - A.y) / (C.y - A.y);
					pd.interpolate(pa,pc,f);
					nd.interpolate(na,nc,f);
					D.x = pd.x / pd.z * dFocalLength + iCenterX;
					D.y = -pd.y / pd.z * dFocalLength + iCenterY;
					if (B.x < D.x) {
						listTriangles.add(new Triangle(A.x,A.y,B.x,B.y,D.x,D.y,pa,pb,pd,na,nb,nd,mp));
						listTriangles.add(new Triangle(C.x,C.y,B.x,B.y,D.x,D.y,pc,pb,pd,nc,nb,nd,mp));
					} else {
						listTriangles.add(new Triangle(A.x,A.y,D.x,D.y,B.x,B.y,pa,pd,pb,na,nd,nb,mp));
						listTriangles.add(new Triangle(C.x,C.y,D.x,D.y,B.x,B.y,pc,pd,pb,nc,nd,nb,mp));
					}
				} else if (B.y > C.y) {
					//System.out.println("..C < B");
					double f = (C.y - A.y) / (B.y - A.y);
					pd.interpolate(pa,pb,f);
					nd.interpolate(na,nb,f);
					D.x = pd.x / pd.z * dFocalLength + iCenterX;
					D.y = -pd.y / pd.z * dFocalLength + iCenterY;
					if (C.x < D.x) {
						listTriangles.add(new Triangle(A.x,A.y,C.x,C.y,D.x,D.y,pa,pc,pd,na,nc,nd,mp));
						listTriangles.add(new Triangle(B.x,B.y,C.x,C.y,D.x,D.y,pb,pc,pd,nb,nc,nd,mp));
					} else {
						listTriangles.add(new Triangle(A.x,A.y,D.x,D.y,C.x,C.y,pa,pd,pc,na,nd,nc,mp));
						listTriangles.add(new Triangle(B.x,B.y,D.x,D.y,C.x,C.y,pb,pd,pc,nb,nd,nc,mp));
					}
				} else {
					//System.out.println(".. B = C");
					if (B.x < C.x) {
						listTriangles.add(new Triangle(A.x,A.y,B.x,B.y,C.x,C.y,pa,pb,pc,na,nb,nc,mp));
					} else {
						listTriangles.add(new Triangle(A.x,C.y,B.x,A.y,C.x,B.y,pa,pc,pb,na,nc,nb,mp));
					}
				}
			} else if (B.y < A.y && B.y < C.y) {
				//System.out.println("B < ..");
				if (A.y < C.y) {
					//System.out.println("..A < C");
					double f = (A.y - B.y) / (C.y - B.y);
					pd.interpolate(pb,pc,f);
					nd.interpolate(nb,nc,f);
					D.x = pd.x / pd.z * dFocalLength + iCenterX;
					D.y = -pd.y / pd.z * dFocalLength + iCenterY;
					if (A.x < D.x) {
						listTriangles.add(new Triangle(B.x,B.y,A.x,A.y,D.x,D.y,pb,pa,pd,nb,na,nd,mp));
						listTriangles.add(new Triangle(C.x,C.y,A.x,A.y,D.x,D.y,pc,pa,pd,nc,na,nd,mp));
					} else {
						listTriangles.add(new Triangle(B.x,B.y,D.x,D.y,A.x,A.y,pb,pd,pa,nb,nd,na,mp));
						listTriangles.add(new Triangle(C.x,C.y,D.x,D.y,A.x,A.y,pc,pd,pa,nc,nd,na,mp));
					}
				} else if (A.y > C.y) {
					//System.out.println("..C < A");
					double f = (C.y - B.y) / (A.y - B.y);
					pd.interpolate(pb,pa,f);
					nd.interpolate(nb,na,f);
					D.x = pd.x / pd.z * dFocalLength + iCenterX;
					D.y = -pd.y / pd.z * dFocalLength + iCenterY;
					if (C.x < D.x) {
						listTriangles.add(new Triangle(B.x,B.y,C.x,C.y,D.x,D.y,pb,pc,pd,nb,nc,nd,mp));
						listTriangles.add(new Triangle(A.x,A.y,C.x,C.y,D.x,D.y,pa,pc,pd,na,nc,nd,mp));
					} else {
						listTriangles.add(new Triangle(B.x,B.y,D.x,D.y,C.x,C.y,pb,pd,pc,nb,nd,nc,mp));
						listTriangles.add(new Triangle(A.x,A.y,D.x,D.y,C.x,C.y,pa,pd,pc,na,nd,nc,mp));
					}
				} else {
					//System.out.println("..A = C");
					if (A.x < C.x) {
						listTriangles.add(new Triangle(B.x,B.y,A.x,A.y,C.x,C.y,pb,pa,pc,nb,na,nc,mp));
					} else {
						listTriangles.add(new Triangle(B.x,B.y,C.x,C.y,A.x,A.y,pb,pc,pa,nb,nc,na,mp));
					}
				}
			} else {
				//System.out.println("C < ..");
				if (A.y < B.y) {
					double f = (A.y - C.y) / (B.y - C.y);
					pd.interpolate(pc,pb,f);
					nd.interpolate(nc,nb,f);
					D.x = pd.x / pd.z * dFocalLength + iCenterX;
					D.y = -pd.y / pd.z * dFocalLength + iCenterY;
					if (A.x < D.x) {
						listTriangles.add(new Triangle(C.x,C.y,A.x,A.y,D.x,D.y,pc,pa,pd,nc,na,nd,mp));
						listTriangles.add(new Triangle(B.x,B.y,A.x,A.y,D.x,D.y,pb,pa,pd,nb,na,nd,mp));
					} else {
						listTriangles.add(new Triangle(C.x,C.y,D.x,D.y,A.x,A.y,pc,pd,pa,nc,nd,na,mp));
						listTriangles.add(new Triangle(B.x,B.y,D.x,D.y,A.x,A.y,pb,pd,pa,nb,nd,na,mp));
					}
				} else if (A.y > B.y) {
					double f = (B.y - C.y) / (A.y - C.y);
					pd.interpolate(pc,pa,f);
					nd.interpolate(nc,na,f);
					D.x = pd.x / pd.z * dFocalLength + iCenterX;
					D.y = -pd.y / pd.z * dFocalLength + iCenterY;
					if (B.x < D.x) {
						listTriangles.add(new Triangle(C.x,C.y,B.x,B.y,D.x,D.y,pc,pb,pd,nc,nb,nd,mp));
						listTriangles.add(new Triangle(A.x,A.y,B.x,B.y,D.x,D.y,pa,pb,pd,na,nb,nd,mp));
					} else {
						listTriangles.add(new Triangle(C.x,C.y,D.x,D.y,B.x,B.y,pc,pd,pb,nc,nd,nb,mp));
						listTriangles.add(new Triangle(A.x,A.y,D.x,D.y,B.x,B.y,pa,pd,pb,na,nd,nb,mp));
					}
				} else {
					if (A.x < B.x) {
						listTriangles.add(new Triangle(C.x,C.y,A.x,A.y,B.x,B.y,pc,pa,pb,nc,na,nb,mp));
					} else {
						listTriangles.add(new Triangle(C.x,C.y,B.x,B.y,A.x,A.y,pc,pb,pa,nc,nb,na,mp));
					}
				}
			}
		}
	}
					
	public void paint(Graphics g) {
		g.drawImage(image,0,0,null);
	}
	
	public Color3f shade(double[] shaderParams,MaterialProperties mp) {
		//System.out.println("shade()");
		Point3d point = new Point3d(shaderParams[Triangle.POSITION_X],shaderParams[Triangle.POSITION_Y],1d / shaderParams[Triangle.POSITION_R]);
		Vector3d normal = new Vector3d(shaderParams[Triangle.NORMAL_X],shaderParams[Triangle.NORMAL_Y],shaderParams[Triangle.NORMAL_Z]);
		normal.normalize();
		Vector3d D = new Vector3d(point);
		D.normalize();
		if (D.dot(normal) >= 0) {
			normal.x = -normal.x;
			normal.y = -normal.y;
			normal.z = -normal.z;
		}

		//System.out.println(point + " " + normal);
		Color3f c3Color = new Color3f(c3AmbientLight);
		c3Color.scale(mp.ambient);
		c3Color.x *= mp.red;
		c3Color.y *= mp.green;
		c3Color.z *= mp.blue;
		Vector3d LV = new Vector3d();
		//System.out.println("p= " + point);
		Vector3d P = new Vector3d(point);
		P.normalize();
		double specular = 0;
		for (int l = 0; l < iLights; l++) {
			if (mp.diffuse != 0) {
				//System.out.println(normal);
				//System.out.println(av3LightDirection[l]);
				//System.out.println(av3LightDirection[l].dot(normal));
				Color3f c3Diffuse = new Color3f(1,1,1);
				c3Diffuse.scale((float) av3LightDirection[l].dot(normal));
				c3Diffuse.clamp(0,1);
				c3Diffuse.x *= ac3LightColor[l].x * mp.red;
				c3Diffuse.y *= ac3LightColor[l].y * mp.green;
				c3Diffuse.z *= ac3LightColor[l].z * mp.blue;
				c3Color.add(c3Diffuse);
			}
			if (abLightSpecular[l] && mp.specular != 0) {
				LV.sub(av3LightDirection[l],P);
				LV.scale(1d/Math.sqrt(LV.dot(LV)));
				LV.normalize();
				double spec = LV.dot(normal);
				Color3f c3Specular = new Color3f(ac3LightColor[l]);
				c3Specular.scale(Math.max((float)Math.pow(spec,1f/mp.roughness) * mp.specular,0));
				c3Specular.clamp(0,1000);
				//System.out.println(c3Specular);
				c3Color.add(c3Specular);
			}
		}
                
		
		//c3Color = new Color3f((float) normal.x * 0.5f + 0.5f,(float) normal.y * 0.5f + 0.5f,(float) normal.z * 0.5f + 0.5f);
		c3Color.clamp(0,1);
		//float red = c3Color.x * 255;
		//float green = c3Color.y * 255;
		//float blue = c3Color.z * 255;
		////red = 128;
		//int color = 0xFF000000 | (((int)red) << 16) | (((int)green) << 8) | ((int)blue);
		//System.out.println(red + " " + green + " " + blue);
		return c3Color;
	}
	
	//public void copyScanlineToBuffer(Scanline scanline, int y) {
	//	int index = y * iWidth;
	//	for (int x = 0; x < iWidth; x++) {
	//		aiFrameBuffer[index++] = scanline.aiColor[x];
	//		//aiFrameBuffer[index++] = 0xFFFF0000;
	//	}
	//}
	
	public void copyScanlinesToBuffer(Scanline[] scanline, int y) {
		int index = y * iWidth;
		Color3f c = new Color3f();
		for (int x = 0; x < iWidth; x++) {
			c.set(0,0,0);
			for (int xx = 0; xx < iSupersample; xx++) {
				for (int yy = 0; yy < iSupersample; yy++) {
					//System.out.println(scanline[yy].ac3Color[x * iSupersample + xx]);
					c.add(scanline[yy].ac3Color[x * iSupersample + xx]);
				}
			}
			//System.out.println("=" + c);
			c.scale(1f / (float) (iSupersample * iSupersample));
			c.clamp(0,1);
			//System.out.println(c);
			float red = c.x * 255;
			float green = c.y * 255;
			float blue = c.z * 255;
			aiFrameBuffer[index++] = 0xFF000000 | (((int)red) << 16) | (((int)green) << 8) | ((int)blue);
		}
	}
	
	public void render() {
		long time = System.currentTimeMillis();
		long start = System.currentTimeMillis();
		Scanline[] scanline = new Scanline[iSupersample];
		for (int y = 0; y < iHeight; y++) {
			for (int s = 0; s < iSupersample; s++) {
				scanline[s] = new Scanline(iWidth * iSupersample, 0xFF000000);
				for (Iterator it = listTriangles.iterator(); it.hasNext(); ) {
					((Triangle) it.next()).fillScanline(scanline[s],y * iSupersample + s);
				}
				scanline[s].shadeScanline(this);
			}
			copyScanlinesToBuffer(scanline,y);
			long newTime = System.currentTimeMillis();
			if (newTime > time + 1000) {
				time = newTime;
				repaint();
			}
		}
		Graphics g = image.getGraphics();
		long deltaT = System.currentTimeMillis() - start;
		float t = (float) deltaT / 1000f;
		g.drawString(iWidth + "x" + iHeight + " (" + iSupersample + "x" + iSupersample + " supersampled)",8,16);
		g.drawString(listTriangles.size() + "  visible triangles",8,32);
		g.drawString("rendered in " + t + " seconds",8,48);
		repaint();
	}
	
	public static void main(String[] args) {
		int width = 640;
		int height = 480;
		int oversample = 4;
		Triangle.init(width * oversample,height * oversample);
		Render2 renderer = new Render2(width,height,oversample);
		JFrame frame = new JFrame();
		frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		renderer.setPreferredSize(new Dimension(width,height));
		frame.getContentPane().add(renderer);
		frame.pack();
		frame.show();
		System.out.println("parsing file " + args[0]);
		renderer.initFromFile(args[0]);
		System.out.println(renderer.listTriangles.size() + " triangles");
		//renderer.addTriangle(
		//	new Point3d(-10,0,50), new Point3d(10,5,50), new Point3d(10,-5,50),
		//	new Vector3d(-1,0,-1), new Vector3d(0,1,-1), new Vector3d(0,-1,-1),
		//	new MaterialProperties()
		//);
		//System.out.println(renderer.listTriangles.size());
		
		System.out.println("starting renderer...");
		renderer.render();
		try {
			ImageIO.write(renderer.image,"PNG",new File(args[1]));
		}
		catch(Exception e) {
			System.err.println(e);
		}
	}
}

