package net.sf.jpatch.boundary;

import javax.vecmath.*;
import net.sf.jpatch.entity.*;
import net.sf.jpatch.auxilary.*;

public final class BoneRepresentation
implements Comparable {
	private final Bone bone;
	private Point3f[] ap3;
	private Point3f p3Center = new Point3f();
	
	public BoneRepresentation(Bone bone) {
		this.bone = bone;
		ap3 = new Point3f[] {
			new Point3f(),
			new Point3f(),
			new Point3f(),
			new Point3f(),
			new Point3f(),
			new Point3f()
		};
	}
	
	public Bone getBone() {
		return bone;
	}
	
	private void compute() {
		Point3f p3Helper = new Point3f();
		ap3[0].set(bone.getStart());
		ap3[5].set(bone.getEnd());
		Vector3f v3Extent = bone.getExtent();
		if (v3Extent.x != 0 || v3Extent.y != 0 || v3Extent.z != 0) {
			Vector3f v3Helper = (v3Extent.x == 0) ? new Vector3f(1,0,0) : new Vector3f(0,1,0);
			v3Helper.cross(v3Helper,v3Extent);
			v3Helper.normalize();
			float radius = v3Extent.length() * 0.075f;
			v3Helper.scale(radius);
			p3Helper.interpolate(ap3[0],ap3[5],0.1f);
			AxisAngle4f axisAngle = new AxisAngle4f(v3Extent,(float)Math.PI/2f);
			Matrix3f m3Rotation = new Matrix3f();
			m3Rotation.set(axisAngle);
			for (int p = 1; p < 5; p++) {
				ap3[p].set(p3Helper);
				ap3[p].add(v3Helper);
				m3Rotation.transform(v3Helper);
			}
		} else {
			for (int p = 1; p < 5; p++) {
				ap3[p].set(ap3[0]);
			}
		}
	}
	
	public Point3f[] getPoints(boolean compute) {
		if (compute) {
			compute();
		}
		return ap3;
	}
	
	public int compareTo(Object o) {
		BoneRepresentation boneRep = (BoneRepresentation)o;
		p3Center.interpolate(ap3[0],ap3[5],0.5f);
		return Float.compare(p3Center.z,Utils3D.closestPointOnLine(boneRep.ap3[0],boneRep.ap3[5],p3Center).z);
	}
	/*	
	private class BoneComparator implements Comparator {
			private Point3f m = new Point3f();
			//private static Point3f p = new Point3f();
			public boolean equals(Object o) {
				return false;
			}
			public int compare(Object o1, Object o2) {
				Point3f[] a = (Point3f[])o1;
				Point3f[] b = (Point3f[])o2;
				m.interpolate(a[0],a[5],0.5f);
				return Float.compare(m.z,Utils3D.closestPointOnLine(b[0],b[5],m).z);
			}
		}
	*/
}
