package net.sf.jpatch.boundary;

import java.awt.image.*;
import java.util.*;
import javax.vecmath.*;

import net.sf.jpatch.auxilary.*;
import net.sf.jpatch.entity.*;

public class IZBufferRenderer {
	private int iWidth;
	private int iHeight;
	private int[] aiFrameBuffer;
	private double[] adIZBuffer;
	private BufferedImage image;

	private Matrix4d m4View;
	private Vector3d[] av3LightDirection;
	private Color3f[] ac3LightColor;
	private boolean[] abLightSpecular;
	private Color3f c3AmbientLight;
	private int iLights;
	private int iLightNum;

	private final double dClipping = 1;
	private double dFocalLength;
	
	public IZBufferRenderer(int width, int height, int backgroundColor) {
		iWidth = width;
		iHeight = height;
		image = new BufferedImage(width, height, BufferedImage.TYPE_INT_RGB);
		aiFrameBuffer = ((DataBufferInt) image.getRaster().getDataBuffer()).getData();
		adIZBuffer = new double[width * height];
		Arrays.fill(aiFrameBuffer, backgroundColor);
		Arrays.fill(adIZBuffer, 0);
	}

	public void setCamera(Matrix4d matrix, double focalLength) {
		m4View = matrix;
		dFocalLength = focalLength;
	}

	public void setNumberOfLights(int lightSources) {
		iLights = lightSources;
		av3LightDirection = new Vector3d[iLights];
		ac3LightColor = new Color3f[iLights];
		abLightSpecular = new boolean[iLights];
		iLightNum = 0;
	}

	public void setAmbientLightColor(Color3f color) {
		c3AmbientLight = color;
	}

	public void addLightSource(Vector3d direction, Color3f color, boolean specular) {
		av3LightDirection[iLightNum] = direction;
		m4View.transform(av3LightDirection[iLightNum]);
		av3LightDirection[iLightNum].normalize();
		ac3LightColor[iLightNum] = color;
		abLightSpecular[iLightNum] = specular;
		iLightNum++;
	}

	public BufferedImage getImage() {
		return image;
	}
	
	public final void transform(Point3d p3, Vector3d v3) {
		m4View.transform(p3);
		Vector3d D = new Vector3d(p3);
		p3.x /= p3.z;
		p3.y /= p3.z;
		p3.x *= dFocalLength;
		p3.y *= dFocalLength;
		p3.z *= dFocalLength;
		m4View.transform(v3);
		D.normalize();
		if (D.dot(v3) >= 0) {
			v3.x = -v3.x;
			v3.y = -v3.y;
			v3.z = -v3.z;
		}
	}
	
	public final void draw3DTrianglePhong(Point3d p1, Point3d p2, Point3d p3, Vector3d n1, Vector3d n2, Vector3d n3, MaterialProperties mp) {
		int x1, y1, x2, y2, x3, y3;
		double nx1, ny1, nz1, nx2, ny2, nz2, nx3, ny3, nz3;
		double px1, py1, pz1, px2, py2, pz2, px3, py3, pz3;
		int dx1, dx2, dy1, dy2, mx1, mx2;
		double dnx1, dnx2, dny1, dny2, dnz1, dnz2, mnx1, mnx2, mny1, mny2, mnz1, mnz2;
		double dpx1, dpx2, dpy1, dpy2, dpz1, dpz2, mpx1, mpx2, mpy1, mpy2, mpz1, mpz2;
		int xstart, xend, yend, y, left, right, i, index;
		double nxstart, nxend, nystart, nyend, nzstart, nzend, nx, ny, nz, dnx, dny, dnz;
		double pxstart, pxend, pystart, pyend, pzstart, pzend, px, py, pz, dpx, dpy, dpz;
		Vector3d normal = new Vector3d();
		Point3d point = new Point3d();
		//Vector3d a = new Vector3d(p2);
		//a.sub(p1);
		//Vector3d b = new Vector3d(p3);
		//b.sub(p1);
		//Vector3d n = new Vector3d();
		//n.cross(a,b);
		//if (n.z > 0) {
		//	n1.scale(-1);
		//	n2.scale(-1);
		//	n3.scale(-1);
		//}
		if (p1.z > dClipping && p2.z > dClipping && p3.z > dClipping) {
			if (p1.y >= p2.y && p1.y >= p3.y) {
				x1 = ((int) p1.x + iWidth / 2) << 16;
				y1 = (iHeight / 2 - (int) p1.y);
				nx1 = n1.x;
				ny1 = n1.y;
				nz1 = n1.z;
				px1 = p1.x;
				py1 = p1.y;
				pz1 = 1d / p1.z;
				if (p2.y > p3.y) {
					x2 = ((int) p2.x + iWidth / 2) << 16;
					y2 = (iHeight / 2 - (int) p2.y);
					nx2 = n2.x;
					ny2 = n2.y;
					nz2 = n2.z;
					px2 = p2.x;
					py2 = p2.y;
					pz2 = 1d / p2.z;
					x3 = ((int) p3.x + iWidth / 2) << 16;
					y3 = (iHeight / 2 - (int) p3.y);
					nx3 = n3.x;
					ny3 = n3.y;
					nz3 = n3.z;
					px3 = p3.x;
					py3 = p3.y;
					pz3 = 1d / p3.z;
				} else {
					x2 = ((int) p3.x + iWidth / 2) << 16;
					y2 = (iHeight / 2 - (int) p3.y);
					nx2 = n3.x;
					ny2 = n3.y;
					nz2 = n3.z;
					px2 = p3.x;
					py2 = p3.y;
					pz2 = 1d /  p3.z;
					x3 = ((int) p2.x + iWidth / 2) << 16;
					y3 = (iHeight / 2 - (int) p2.y);
					nx3 = n2.x;
					ny3 = n2.y;
					nz3 = n2.z;
					px3 = p2.x;
					py3 = p2.y;
					pz3 = 1d / p2.z;
				}
			} else if (p2.y >= p1.y && p2.y >= p3.y) {
				x1 = ((int) p2.x + iWidth / 2) << 16;
				y1 = (iHeight / 2 - (int) p2.y);
				nx1 = n2.x;
				ny1 = n2.y;
				nz1 = n2.z;
				px1 = p2.x;
				py1 = p2.y;
				pz1 = 1d / p2.z;
				if (p1.y > p3.y) {
					x2 = ((int) p1.x + iWidth / 2) << 16;
					y2 = (iHeight / 2 - (int) p1.y);
					nx2 = n1.x;
					ny2 = n1.y;
					nz2 = n1.z;
					px2 = p1.x;
					py2 = p1.y;
					pz2 = 1d / p1.z;
					x3 = ((int) p3.x + iWidth / 2) << 16;
					y3 = (iHeight / 2 - (int) p3.y);
					nx3 = n3.x;
					ny3 = n3.y;
					nz3 = n3.z;
					px3 = p3.x;
					py3 = p3.y;
					pz3 = 1d / p3.z;
				} else {
					x2 = ((int) p3.x + iWidth / 2) << 16;
					y2 = (iHeight / 2 - (int) p3.y);
					nx2 = n3.x;
					ny2 = n3.y;
					nz2 = n3.z;
					px2 = p3.x;
					py2 = p3.y;
					pz2 = 1d / p3.z;
					x3 = ((int) p1.x + iWidth / 2) << 16;
					y3 = (iHeight / 2 - (int) p1.y);
					nx3 = n1.x;
					ny3 = n1.y;
					nz3 = n1.z;
					px3 = p1.x;
					py3 = p1.y;
					pz3 = 1d / p1.z;
				}
			} else {
				x1 = ((int) p3.x + iWidth / 2) << 16;
				y1 = (iHeight / 2 - (int) p3.y);
				nx1 = n3.x;
				ny1 = n3.y;
				nz1 = n3.z;
				px1 = p3.x;
				py1 = p3.y;
				pz1 = 1d / p3.z;
				if (p1.y > p2.y) {
					x2 = ((int) p1.x + iWidth / 2) << 16;
					y2 = (iHeight / 2 - (int) p1.y);
					nx2 = n1.x;
					ny2 = n1.y;
					nz2 = n1.z;
					px2 = p1.x;
					py2 = p1.y;
					pz2 = 1d / p1.z;
					x3 = ((int) p2.x + iWidth / 2) << 16;
					y3 = (iHeight / 2 - (int) p2.y);
					nx3 = n2.x;
					ny3 = n2.y;
					nz3 = n2.z;
					px3 = p2.x;
					py3 = p2.y;
					pz3 = 1d / p2.z;
				} else {
					x2 = ((int) p2.x + iWidth / 2) << 16;
					y2 = (iHeight / 2 - (int) p2.y);
					nx2 = n2.x;
					ny2 = n2.y;
					nz2 = n2.z;
					px2 = p2.x;
					py2 = p2.y;
					pz2 = 1d / p2.z;
					x3 = ((int) p1.x + iWidth / 2) << 16;
					y3 = (iHeight / 2 - (int) p1.y);
					nx3 = n1.x;
					ny3 = n1.y;
					nz3 = n1.z;
					px3 = p1.x;
					py3 = p1.y;
					pz3 = 1d / p1.z;
				}
			}
			//System.out.println("" + (x1 >> 16) + " " + y1 + " " + (x2 >> 16) + " " + y2 + " " + (x3 >> 16) + " " + y3);
			//System.out.println("" + px1 + " " + py1 + " " + 1d / pz1 + " " + px2 + " " + py2 + " " + 1d / pz2 + " " + px3 + " " + py3 + " " + 1d / pz3);
			dx1 = x3 - x1;
			dy1 = y3 - y1;
			dnx1 = nx3 - nx1;
			dny1 = ny3 - ny1;
			dnz1 = nz3 - nz1;
			dpx1 = px3 - px1;
			dpy1 = py3 - py1;
			dpz1 = pz3 - pz1;
			if (dy1 == 0) {
				return;
			}
			dx2 = x2 - x1;
			dy2 = y2 - y1;
			dnx2 = nx2 - nx1;
			dny2 = ny2 - ny1;
			dnz2 = nz2 - nz1;
			dpx2 = px2 - px1;
			dpy2 = py2 - py1;
			dpz2 = pz2 - pz1;
			mx1 = dx1 / dy1;
			mnx1 = dnx1 / dy1;
			mny1 = dny1 / dy1;
			mnz1 = dnz1 / dy1;
			mpx1 = dpx1 / dy1;
			mpy1 = dpy1 / dy1;
			mpz1 = dpz1 / dy1;
			xstart = xend = x1;
			nxstart = nxend = nx1;
			nystart = nyend = ny1;
			nzstart = nzend = nz1;
			pxstart = pxend = px1;
			pystart = pyend = py1;
			pzstart = pzend = pz1;
			y = y1;
			if (dy2 != 0) {
				mx2 = dx2 / dy2;
				mnx2 = dnx2 / dy2;
				mny2 = dny2 / dy2;
				mnz2 = dnz2 / dy2;
				mpx2 = dpx2 / dy2;
				mpy2 = dpy2 / dy2;
				mpz2 = dpz2 / dy2;
				if (y2 < 0) {
					xstart += mx1 * dy2;
					xend += mx2 * dy2;
					nxstart += nx1 * dy2;
					nxend += nx2 * dy2;
					nystart += ny1 * dy2;
					nyend += ny2 * dy2;
					nzstart += nz1 * dy2;
					nzend += nz2 * dy2;
					pxstart += px1 * dy2;
					pxend += px2 * dy2;
					pystart += py1 * dy2;
					pyend += py2 * dy2;
					pzstart += pz1 * dy2;
					pzend += pz2 * dy2;
					y = y2;
				} else if (y < 0) {
					xstart -= mx1 * y;
					xend -= mx2 * y;
					nxstart -= nx1 * y;
					nxend -= nx2 * y;
					nystart -= ny1 * y;
					nyend -= ny2 * y;
					nzstart -= nz1 * y;
					nzend -= nz2 * y;
					pxstart -= px1 * y;
					pxend -= px2 * y;
					pystart -= py1 * y;
					pyend -= py2 * y;
					pzstart -= pz1 * y;
					pzend -= pz2 * y;
					y = 0;
				}
				yend = (y2 < iHeight) ? y2 : iHeight;
				index = y * iWidth;
				while (y < yend) {
					if (xstart < xend) {
						left = xstart >> 16;
						right = xend >> 16;
						nx = nxstart;
						dnx = nxend - nxstart;
						ny = nystart;
						dny = nyend - nystart;
						nz = nzstart;
						dnz = nzend - nzstart;
						px = pxstart;
						dpx = pxend - pxstart;
						py = pystart;
						dpy = pyend - pystart;
						pz = pzstart;
						dpz = pzend - pzstart;
					} else {
						left = xend >> 16;
						right = xstart >> 16;
						nx = nxend;
						dnx = nxstart - nxend;
						ny = nyend;
						dny = nystart - nyend;
						nz = nzend;
						dnz = nzstart - nzend;
						px = pxend;
						dpx = pxstart - pxend;
						py = pyend;
						dpy = pystart - pyend;
						pz = pzend;
						dpz = pzstart - pzend;
					}
					if (left != right) {
						double div = (float) (right - left);
						dnx /= div;
						dny /= div;
						dnz /= div;
						dpx /= div;
						dpy /= div;
						dpz /= div;
						if (left < 0) {
							double l = (double) left;
							nx -= l * dnx;
							ny -= l * dny;
							nz -= l * dnz;
							px -= l * dpx;
							py -= l * dpy;
							pz -= l * dpz;
							left = 0;
						}
						if (right > iWidth) {
							right = iWidth;
						}
						for (i = left; i < right; i++) {
							if (pz > adIZBuffer[index + i]) {
								normal.set(nx,ny,nz);
								point.set(px,py,1d / pz);
								//Vector3d D = new Vector3d(point);
								//D.normalize();
								//if (D.dot(normal) > 0) {
								//	normal.x = -normal.x;
								//	normal.y = -normal.y;
								//	normal.z = -normal.z;
								//}
								aiFrameBuffer[index + i] = shade(point,normal,mp);
								adIZBuffer[index + i] = pz;
							}
							nx += dnx;
							ny += dny;
							nz += dnz;
							px += dpx;
							py += dpy;
							pz += dpz;
						}
					}
					xstart += mx1;
					nxstart += mnx1;
					nystart += mny1;
					nzstart += mnz1;
					pxstart += mpx1;
					pystart += mpy1;
					pzstart += mpz1;
					xend += mx2;
					nxend += mnx2;
					nyend += mny2;
					nzend += mnz2;
					pxend += mpx2;
					pyend += mpy2;
					pzend += mpz2;
					index += iWidth;
					y++;
				}
			}
			dx2 = x3 - x2;
			dy2 = y3 - y2;
			dnx2 = nx3 - nx2;
			dny2 = ny3 - ny2;
			dnz2 = nz3 - nz2;
			dpx2 = px3 - px2;
			dpy2 = py3 - py2;
			dpz2 = pz3 - pz2;
			if (dy2 != 0) {
				mx2 = dx2 / dy2;
				mnx2 = dnx2 / dy2;
				mny2 = dny2 / dy2;
				mnz2 = dnz2 / dy2;
				mpx2 = dpx2 / dy2;
				mpy2 = dpy2 / dy2;
				mpz2 = dpz2 / dy2;
				xend = x2;
				nxend = nx2;
				nyend = ny2;
				nzend = nz2;
				pxend = px2;
				pyend = py2;
				pzend = pz2;
				if (y < 0) {
					xstart -= mx1 * y;
					xend -= mx2 * y;
					nxstart -= mnx1 * y;
					nxend -= mnx2 * y;
					nystart -= mny1 * y;
					nyend -= mny2 * y;
					nzstart -= mnz1 * y;
					nzend -= mnz2 * y;
					pxstart -= mpx1 * y;
					pxend -= mpx2 * y;
					pystart -= mpy1 * y;
					pyend -= mpy2 * y;
					pzstart -= mpz1 * y;
					pzend -= mpz2 * y;
					y = 0;
				}
				yend = (y3 < iHeight) ? y3 : iHeight;
				index = y * iWidth;
				while (y < yend) {
					if (xstart < xend) {
						left = xstart >> 16;
						right = xend >> 16;
						nx = nxstart;
						dnx = nxend - nxstart;
						ny = nystart;
						dny = nyend - nystart;
						nz = nzstart;
						dnz = nzend - nzstart;
						px = pxstart;
						dpx = pxend - pxstart;
						py = pystart;
						dpy = pyend - pystart;
						pz = pzstart;
						dpz = pzend - pzstart;
					} else {
						left = xend >> 16;
						right = xstart >> 16;
						nx = nxend;
						dnx = nxstart - nxend;
						ny = nyend;
						dny = nystart - nyend;
						nz = nzend;
						dnz = nzstart - nzend;
						px = pxend;
						dpx = pxstart - pxend;
						py = pyend;
						dpy = pystart - pyend;
						pz = pzend;
						dpz = pzstart - pzend;
					}
					if (left != right) {
						double div = (double) (right - left);
						dnx /= div;
						dny /= div;
						dnz /= div;
						dpx /= div;
						dpy /= div;
						dpz /= div;
						if (left < 0) {
							double l = (double) left;
							nx -= l * dnx;
							ny -= l * dny;
							nz -= l * dnz;
							px -= l * dpx;
							py -= l * dpy;
							pz -= l * dpz;
							left = 0;
						}
						if (right > iWidth) {
							right = iWidth;
						}
						for (i = left; i < right; i++) {
							if (pz > adIZBuffer[index + i]) {
								normal.set(nx,ny,nz);
								point.set(px,py,1d / pz);
								//Vector3d D = new Vector3d(point);
								//D.normalize();
								//if (D.dot(normal) > 0) {
								//	normal.x = -normal.x;
								//	normal.y = -normal.y;
								//	normal.z = -normal.z;
								//}
								aiFrameBuffer[index + i] = shade(point,normal,mp);
								adIZBuffer[index + i] = pz;
							}
							nx += dnx;
							ny += dny;
							nz += dnz;
							px += dpx;
							py += dpy;
							pz += dpz;
						}
					}
					xstart += mx1;
					nxstart += mnx1;
					nystart += mny1;
					nzstart += mnz1;
					pxstart += mpx1;
					pystart += mpy1;
					pzstart += mpz1;
					xend += mx2;
					nxend += mnx2;
					nyend += mny2;
					nzend += mnz2;
					pxend += mpx2;
					pyend += mpy2;
					pzend += mpz2;
					index += iWidth;
					y++;
				}
			}
		}
	}
	
	private final int shade(Point3d point, Vector3d normal, MaterialProperties mp) {
		Color3f c3Color = new Color3f(c3AmbientLight);
		c3Color.scale(mp.ambient);
		c3Color.x *= mp.red;
		c3Color.y *= mp.green;
		c3Color.z *= mp.blue;
		Vector3d LV = new Vector3d();
		//System.out.println("p= " + point);
		Vector3d P = new Vector3d(point);
		P.normalize();
		double specular = 0;
		for (int l = 0; l < iLights; l++) {
			if (mp.diffuse != 0) {
				//System.out.println(normal);
				//System.out.println(av3LightDirection[l]);
				//System.out.println(av3LightDirection[l].dot(normal));
				Color3f c3Diffuse = new Color3f(1,1,1);
				c3Diffuse.scale((float) av3LightDirection[l].dot(normal));
				c3Diffuse.clamp(0,1);
				c3Diffuse.x *= ac3LightColor[l].x * mp.red;
				c3Diffuse.y *= ac3LightColor[l].y * mp.green;
				c3Diffuse.z *= ac3LightColor[l].z * mp.blue;
				c3Color.add(c3Diffuse);
			}
			if (abLightSpecular[l] && mp.specular != 0) {
				LV.sub(av3LightDirection[l],P);
				LV.scale(1d/Math.sqrt(LV.dot(LV)));
				LV.normalize();
				double spec = LV.dot(normal);
				Color3f c3Specular = new Color3f(ac3LightColor[l]);
				c3Specular.scale(Math.max((float)Math.pow(spec,1f/mp.roughness) * mp.specular,0));
				c3Specular.clamp(0,1000);
				//System.out.println(c3Specular);
				c3Color.add(c3Specular);
			}
		}

		//float red = Math.min(mp.red * diffuse + specular, 1) * 255;
		//float green = Math.min(mp.green * diffuse + specular, 1) * 255;
		//float blue = Math.min(mp.blue * diffuse + specular, 1) * 255;

		c3Color.clamp(0,1);
		float red = c3Color.x * 255;
		float green = c3Color.y * 255;
		float blue = c3Color.z * 255;
		
		int color = 0xFF000000 | (((int)red) << 16) | (((int)green) << 8) | ((int)blue);
		//System.out.println(red + " " + green + " " + blue);
		return color;
	}
}

