package net.sf.jpatch.boundary;

import java.util.*;
import java.awt.*;
import java.awt.image.*;
import java.awt.geom.*;
import javax.swing.*;
import javax.vecmath.*;
import net.sf.jpatch.renderer.*;
import net.sf.jpatch.entity.*;
import net.sf.jpatch.auxilary.*;
import net.sf.jpatch.boundary.selection.*;
import net.sf.jpatch.boundary.tools.*;

public class JPatchCanvas extends JPanel
implements Viewport {
	
	private static final int GHOST = JPatchSettings.getInstance().iGhost;
	
	private static int COUNT = 0;
	private static int NUM = 0;
	private int num;
	
	private static final int EVERYTHING = 0;
	private static final int BACKGROUND = 1;
	private static final int FOREGROUND = 2;
	
	private static int count = 0;

	private int iMode = EVERYTHING;
	
	//private Model model;
	private ViewDefinition viewDefinition;
	private JPatchDrawable drawable;
	private Image image;
	private Matrix4f m4View;
	private boolean bZBuffer = false;
	private int[] aiStaticFrameBuffer;
	private int[] aiStaticZBuffer;
	private JPatchTool tool;
	private Grid grid = new Grid();
	private boolean bFlat = true;
	private Lighting lighting;
	private JPatchSettings settings = JPatchSettings.getInstance();
	
	private Font font = new Font("SansSerif",Font.PLAIN,12);
	
	private ArrayList listTransparentPatches = new ArrayList();
	// remove !//
	//Rotoscope rotoscope = new Rotoscope("f:/zbuffer.png");
	//
	
	public JPatchCanvas(Model model, ViewDefinition viewDefinition) {
		//this.model = model;
		this.viewDefinition = viewDefinition;
		viewDefinition.setViewport(this);
		m4View = viewDefinition.getMatrix();
		num = NUM ++;
	}

	public void updateImage() {
		if (viewDefinition.renderPatches() || viewDefinition.alwaysUseZBuffer()) {
			drawable = new JPatchDrawableZBuffer(getGraphicsConfiguration().createCompatibleImage(getWidth(), getHeight()), lighting);
			aiStaticFrameBuffer = new int[getWidth() * getHeight()];
			aiStaticZBuffer = new int[getWidth() * getHeight()];
			bZBuffer = true;
		} else {
			drawable = new JPatchDrawableJava2D(createVolatileImage(getWidth(), getHeight()));
			bZBuffer = false;
		}
		image = drawable.getImage();
		render();
	}
	
	public ViewDefinition getViewDefinition() {
		return viewDefinition;
	}
	
	public JPatchDrawable getDrawable() {
		return drawable;
	}
	
	public Grid getGrid() {
		return grid;
	}
	
	public void flatShade(boolean flat) {
		bFlat = flat;
	}
	
	public void setLighting(Lighting lighting) {
		this.lighting = lighting;
		if (bZBuffer) {
			((JPatchDrawableZBuffer) drawable).setLighting(lighting);
		}
	}
	
	public Lighting getLighting() {
		return lighting;
	}
	
	public void prepareBackground() {
		if (viewDefinition.renderPatches()) {
			PointSelection ps = MainFrame.getInstance().getPointSelection();
			if (ps != null) {
				for (Patch patch = MainFrame.getInstance().getModel().getFirstPatch(); patch != null; patch = patch.getNext()) {
					patch.check(ps);
				}
			}
			((ZBufferRenderer)drawable).setActiveBuffers(aiStaticFrameBuffer, aiStaticZBuffer);
			drawable.clear();
			renderPatches(BACKGROUND);
			iMode = FOREGROUND;
			((ZBufferRenderer)drawable).renderToImage();
		}
	}
	
	public void clearBackground() {
		iMode = EVERYTHING;
		;
	}
	
	public void setTool(JPatchTool tool) {
		
		//tool = new RotoscopeTool(rotoscope);
		this.tool = tool;
		//System.out.println("set tool " + tool);
		if (tool != null) {
			addMouseListener(tool);
		}
		repaint();
	}
	
	//public void updateComponent(Graphics g) {
	//	//if (image == null || image.getWidth(this) != getWidth() || image.getHeight(this) != getHeight()) {
	//	//	updateImage();
	//	//}
	//	//g.drawImage(image, 0, 0, this);
	//}
	
	public void clearImage() {
		image = null;
		viewDefinition.computeMatrix();
	}
	
	public void paintComponent(Graphics g) {
		//super.paintComponent(g);
		//System.out.print(COUNT++ + " JPatchCanvas " + num + " paintComponent() " + getWidth() + "x" + getHeight());
		if (image == null || image.getWidth(this) != getWidth() || image.getHeight(this) != getHeight()) {
		//	System.out.print(" updateImage()");
			updateImage();
		}
		//System.out.println();
		g.drawImage(image, 0, 0, this);
	}

	private void drawRotoscope() {
		Rotoscope rotoscope = MainFrame.getInstance().getModel().getRotoscope(viewDefinition.getView());
		if (rotoscope != null) {
			rotoscope.paint(this);
		}
	}
		
		
	private void drawGrid() {
		grid.paint(this);
		Graphics2D g2 = (Graphics2D)drawable.getGraphics();
		g2.setColor(settings.cBackground);
		g2.fillOval(-5,getHeight() - 85,90,90);
		g2.setColor(settings.cGridMin);
		g2.drawOval(-5,getHeight() - 85,90,90);
	}
	
	public void drawActiveViewportMarker(boolean active) {
		Graphics2D g2 = (Graphics2D)drawable.getGraphics();
		if (active) {
			g2.setColor(settings.cSelection);
		} else {
			g2.setColor(settings.cBackground);
		}
		g2.drawRect(0,0,getWidth() - 1,getHeight() - 1);
	}
	
	private void drawDisplayName() {
		Graphics2D g2 = (Graphics2D)drawable.getGraphics();
		g2.setColor(settings.cText);
		g2.setFont(font);
		g2.drawString(viewDefinition.getViewName(),7,15);
		if (bZBuffer) {
			g2.drawLine(getWidth() - 13,7,getWidth() - 6, 7);
			g2.drawLine(getWidth() - 13,14,getWidth() - 6, 14);
			g2.drawLine(getWidth() - 13,14,getWidth() - 6, 7);
		}
		if (grid.isSnapping()) {
			g2.drawLine(getWidth() - 8,20,getWidth() - 8,29);
			g2.drawLine(getWidth() - 11,20,getWidth() - 11,29);
			g2.drawLine(getWidth() - 5,23,getWidth() - 14,23);
			g2.drawLine(getWidth() - 5,26,getWidth() - 14,26);
		}
	}
	
	private void drawOrigin() {
		float len = 16f/m4View.getScale();;
		Point3f p1 = new Point3f();
		Point3f p2 = new Point3f();
		p1.set(len,0,0);
		p2.set(-len,0,0);
		m4View.transform(p1);
		m4View.transform(p2);
		//p1.z -= 10000f;
		//p2.z -= 10000f;
		drawable.setColor(settings.cX);
		drawable.drawGhostLine3D(p1,p2,GHOST);
		p1.set(0,len,0);
		p2.set(0,-len,0);
		m4View.transform(p1);
		m4View.transform(p2);
		//p1.z -= 10000f;
		//p2.z -= 10000f;
		drawable.setColor(settings.cY);
		drawable.drawGhostLine3D(p1,p2,GHOST);
		p1.set(0,0,len);
		p2.set(0,0,-len);
		m4View.transform(p1);
		m4View.transform(p2);
		//p1.z -= 10000f;
		//p2.z -= 10000f;
		drawable.setColor(settings.cZ);
		drawable.drawGhostLine3D(p1,p2,GHOST);
	}
	
	private void renderPatches(int mode) {
		if (!bFlat) {
			lighting.setRotation(viewDefinition.getRotateX(),viewDefinition.getRotateY());
			lighting.transform();
		}
		//switch (mode) {
		listTransparentPatches.clear();
		
		/* opaque pass */
		switch (EVERYTHING) {
			case EVERYTHING:
				for (Patch patch = MainFrame.getInstance().getModel().getFirstPatch(); patch != null; patch = patch.getNext()) {
					if (!patch.isHidden()) {
						MaterialProperties materialProperties = patch.getMaterial().getMaterialProperties();
						if (materialProperties.isOpaque()) {
							Point3f[][] bicubicPatches = patch.bicubicPatches();
							for (int p = 0; p < bicubicPatches.length; p++) {
								Point3f[] bezierCPs = bicubicPatches[p];
								for (int i = 0; i < bezierCPs.length; i++) {
									m4View.transform(bezierCPs[i]);
								}
								if (bFlat) {
									drawable.drawBicubicPatchFlat(bezierCPs,materialProperties);
								} else {
									drawable.drawBicubicPatchGourad(bezierCPs,materialProperties);
								}
							}
						} else {
							listTransparentPatches.add(patch);
						}
					}
				}
			break;
			case BACKGROUND:
				drawGrid();
				for (Patch patch = MainFrame.getInstance().getModel().getFirstPatch(); patch != null; patch = patch.getNext()) {
					if (!patch.isHidden() && !patch.isDynamic()) {
						MaterialProperties materialProperties = patch.getMaterial().getMaterialProperties();
						if (materialProperties.isOpaque()) {
							Point3f[][] bicubicPatches = patch.bicubicPatches();
							for (int p = 0; p < bicubicPatches.length; p++) {
								Point3f[] bezierCPs = bicubicPatches[p];
								for (int i = 0; i < bezierCPs.length; i++) {
									m4View.transform(bezierCPs[i]);
								}
								if (bFlat) {
									drawable.drawBicubicPatchFlat(bezierCPs,materialProperties);
								} else {
									drawable.drawBicubicPatchGourad(bezierCPs,materialProperties);
								}
							}
						} else {
							listTransparentPatches.add(patch);
						}
					}
				}
			break;
			case FOREGROUND:
				for (Patch patch = MainFrame.getInstance().getModel().getFirstPatch(); patch != null; patch = patch.getNext()) {
					if (!patch.isHidden() && patch.isDynamic()) {
						MaterialProperties materialProperties = patch.getMaterial().getMaterialProperties();
						if (materialProperties.isOpaque()) {
							Point3f[][] bicubicPatches = patch.bicubicPatches();
							for (int p = 0; p < bicubicPatches.length; p++) {
								Point3f[] bezierCPs = bicubicPatches[p];
								for (int i = 0; i < bezierCPs.length; i++) {
									m4View.transform(bezierCPs[i]);
								}
								if (bFlat) {
									drawable.drawBicubicPatchFlat(bezierCPs,materialProperties);
								} else {
									drawable.drawBicubicPatchGourad(bezierCPs,materialProperties);
								}
							}
						} else {
							listTransparentPatches.add(patch);
						}
					}
				}
			break;
		}
		
		/* transparent pass */
		for (Iterator it = listTransparentPatches.iterator(); it.hasNext(); ) {
			Patch patch = (Patch) it.next();
			MaterialProperties materialProperties = patch.getMaterial().getMaterialProperties();
			Point3f[][] bicubicPatches = patch.bicubicPatches();
			for (int p = 0; p < bicubicPatches.length; p++) {
				Point3f[] bezierCPs = bicubicPatches[p];
				for (int i = 0; i < bezierCPs.length; i++) {
					m4View.transform(bezierCPs[i]);
				}
				if (bFlat) {
					drawable.drawBicubicPatchFlat(bezierCPs,materialProperties);
				} else {
					drawable.drawBicubicPatchGourad(bezierCPs,materialProperties);
				}
			}
		}
	}
	
	private void paintTest() {
		/*
		SimpleShape shape = SimpleShape.createCube(0.0f, 1.0f, 1.0f);
		drawable.drawSimpleShape(shape, m4View);
		shape.setColor(1.0f,0.0f,0.0f);
		Matrix4f m4 = new Matrix4f();
		m4.set(new Vector3f(3,0,0));
		shape.transform(m4);
		drawable.drawSimpleShape(shape, m4View);
		*/
	}
	
	private void paintAxis() {
		Matrix3f m3 = new Matrix3f();
		m4View.get(m3);
		Vector3f v3x = new Vector3f(30,0,0);
		Vector3f v3y = new Vector3f(0,30,0);
		Vector3f v3z = new Vector3f(0,0,30);
		m3.transform(v3x);
		m3.transform(v3y);
		m3.transform(v3z);
		int cx = 40;
		int cy = getHeight() - 40;
		int X,Y,Z;
		
		SimpleShape arrowX = SimpleShape.createArrow(9,3,1,0,0);
		SimpleShape arrowY = SimpleShape.createArrow(9,3,0,1,0);
		SimpleShape arrowZ = SimpleShape.createArrow(9,3,0.5f,0.5f,1);
		Matrix3f rot = new Matrix3f();
		rot.rotX((float)Math.PI);
		arrowZ.transform(rot);
		arrowZ.translate(new Vector3f(0,0,31));
		rot.rotX((float)Math.PI / 2);
		arrowY.transform(rot);
		arrowY.translate(new Vector3f(0,31,0));
		rot.rotY(-(float)Math.PI / 2);
		arrowX.transform(rot);
		arrowX.translate(new Vector3f(31,0,0));
		arrowX.transform(m3);
		arrowY.transform(m3);
		arrowZ.transform(m3);
		arrowX.translate(new Vector3f(cx,cy,-10000));
		arrowY.translate(new Vector3f(cx,cy,-10000));
		arrowZ.translate(new Vector3f(cx,cy,-10000));
		
		if (v3x.z > v3y.z && v3y.z > v3z.z) {
			X = 0;
			Y = 1;
			Z = 2;
		} else if (v3x.z > v3z.z && v3z.z > v3y.z) {
			X = 0;
			Z = 1;
			Y = 2;
		} else if (v3y.z > v3x.z && v3x.z > v3z.z) {
			Y = 0;
			X = 1;
			Z = 2;
		} else if (v3z.z > v3x.z && v3x.z > v3y.z) {
			Z = 0;
			X = 1;
			Y = 2;
		} else if (v3y.z > v3z.z && v3z.z > v3x.z) {
			Y = 0;
			Z = 1;
			X = 2;
		} else {
			Z = 0;
			Y = 1;
			X = 2;
		}
		for (int i = 0; i < 3; i++) {
			if (i == X) {
				drawable.setColor(settings.cX);
				drawable.drawLine(cx,cy,cx + (int)v3x.x, cy + (int)v3x.y);
				v3x.scale(1.3f);
				//g2.setColor(new Color(255,64,64));
				drawable.drawLine(cx + (int)v3x.x - 3, cy + (int)v3x.y - 3, cx + (int)v3x.x + 3, cy + (int)v3x.y + 3);
				drawable.drawLine(cx + (int)v3x.x - 3, cy + (int)v3x.y + 3, cx + (int)v3x.x + 3, cy + (int)v3x.y - 3);
				drawable.drawSimpleShape(arrowX,null);
			} else if (i == Y) {
				drawable.setColor(settings.cY);
				drawable.drawLine(cx,cy,cx + (int)v3y.x, cy + (int)v3y.y);
				v3y.scale(1.3f);
				//g2.setColor(new Color(32,255,32));
				drawable.drawLine(cx + (int)v3y.x, cy + (int)v3y.y, cx + (int)v3y.x - 3, cy + (int)v3y.y - 3);
				drawable.drawLine(cx + (int)v3y.x, cy + (int)v3y.y, cx + (int)v3y.x + 3, cy + (int)v3y.y - 3);
				drawable.drawLine(cx + (int)v3y.x, cy + (int)v3y.y, cx + (int)v3y.x, cy + (int)v3y.y + 3);
				drawable.drawSimpleShape(arrowY,null);
			} else {
				drawable.setColor(settings.cZ);
				drawable.drawLine(cx,cy,cx + (int)v3z.x, cy + (int)v3z.y);
				v3z.scale(1.3f);
				//drawable.setColor(new Color(64,64,255));
				drawable.drawLine(cx + (int)v3z.x - 3, cy + (int)v3z.y - 3, cx + (int)v3z.x + 3, cy + (int)v3z.y - 3);
				drawable.drawLine(cx + (int)v3z.x - 3, cy + (int)v3z.y + 3, cx + (int)v3z.x + 3, cy + (int)v3z.y + 3);
				drawable.drawLine(cx + (int)v3z.x - 3, cy + (int)v3z.y + 3, cx + (int)v3z.x + 3, cy + (int)v3z.y - 3);
				drawable.drawSimpleShape(arrowZ,null);
			}
		}
		//g2.setStroke(new BasicStroke(1));
	}
	
	public void render() {
		//System.out.println("\n" + COUNT++ + " JPatchCanvas " + num + " render()");
		Model model = MainFrame.getInstance().getModel();
		//System.out.println("repaint " + count++);
		if (drawable != null && image != null) {
			long start = System.currentTimeMillis();
			if (!bZBuffer || iMode == EVERYTHING) {
				drawable.clear();
				if (viewDefinition.showRotoscope()) {
					drawRotoscope();
				}
				drawGrid();
			} else {
				((ZBufferRenderer)drawable).reset(aiStaticFrameBuffer, aiStaticZBuffer);
			}
			if (iMode != BACKGROUND) {
				if (viewDefinition.renderCurves()) {
					drawable.setColor(settings.cCurve);
					//long start = System.currentTimeMillis();
					//for (int i = 0; i < 100; i++) {
					for(Curve curve = model.getFirstCurve(); curve != null; curve = curve.getNext()) {
						if (!curve.getStart().isStartHook()) {
							drawable.drawJPatchCurve3D(curve, m4View);
						}
					}
					//}
					//System.out.println(System.currentTimeMillis() - start);
				}
				if (viewDefinition.renderPoints()) {
					
					Graphics2D g2 = (Graphics2D)drawable.getGraphics();
					g2.setColor(settings.cText);
					g2.setFont(font);
							
					PointSelection ps = MainFrame.getInstance().getPointSelection();
					Point3f p = new Point3f();
					for(Curve curve = model.getFirstCurve(); curve != null; curve = curve.getNext()) {
						for(ControlPoint cp = curve.getStart(); cp != null; cp = cp.getNextCheckNextLoop()) {
							if (cp.isHead()) {
								p.set(cp.getPosition());
								m4View.transform(p);
								if (ps != null && ps.contains(cp)) {
									drawable.setColor(settings.cPointSelected);
									drawable.drawPoint3D(p,1);
								} else if (!cp.isHook() && ! cp.isHidden()){
									if (cp.isSingle()) {
										drawable.setColor(settings.cPoint);
										drawable.drawXPoint3D(p);
									} else {
										drawable.setColor(settings.cHeadPoint);
										drawable.drawPoint3D(p,1);
									}
									//if (cp.isHidden()) drawable.setColor(Color.BLUE);
									
								}
								//if (cp.getHookPos() != 0 && cp.getHookPos() != 1) {
								//	g2.drawString("" + cp.number(),(int) p.x,(int) p.y);
								//}
								///* render Tangents... */
								//
								//
								//Point3f p3OutTangent = cp.getOutTangent();
								//Point3f p3InTangent = cp.getInTangent();
								//drawable.setColor(settings.cSelection);
								//if (p3OutTangent != null) {
								//	Point3f pt = new Point3f(p3OutTangent);
								//	m4View.transform(pt);
								//	drawable.drawLine3D(p,pt);
								//}
								//if (p3InTangent != null) {
								//	Point3f pt = new Point3f(p3InTangent);
								//	m4View.transform(pt);
								//	drawable.drawLine3D(p,pt);
								//}
								
							}
							//else if (cp.number() == 31 || cp.number() == 29 || cp.number() == 256 || cp.number() == 302) {
							//	p.set(cp.getPosition());
							//	m4View.transform(p);
							//	g2.drawString("" + cp.number(),(int) p.x + 4,(int) p.y + 12);
							//}
							
							
						}
					}
					if (ps != null && ps.getHotCp() != null) {
						if (ps.getHotCp().getCurve() != null && ps.getHotCp().getCurve().getModel() != null) {
							p.set(ps.getHotCp().getPosition());
							m4View.transform(p);
							drawable.setColor(settings.cHot);
							drawable.drawPoint3D(p,2);
						}
					}
					if (ps != null && ps.isCurve()) {
						ControlPoint cp = ps.getControlPoint();
						if (ps.getDirection() && cp.getNext() != null) {
							Point3f p3A = new Point3f(cp.getPosition());
							Point3f p3B = new Point3f(cp.getOutTangent());
							Point3f p3C = new Point3f(cp.getNext().getInTangent());
							Point3f p3D = new Point3f(cp.getNext().getPosition());
							m4View.transform(p3A);
							m4View.transform(p3B);
							m4View.transform(p3C);
							m4View.transform(p3D);
							drawable.setColor(settings.cSelected);
							drawable.drawCurveSegment(p3A,p3B,p3C,p3D);
						} else if (!ps.getDirection() && cp.getPrev() != null) {
							Point3f p3A = new Point3f(cp.getPrev().getPosition());
							Point3f p3B = new Point3f(cp.getPrev().getOutTangent());
							Point3f p3C = new Point3f(cp.getInTangent());
							Point3f p3D = new Point3f(cp.getPosition());
							m4View.transform(p3A);
							m4View.transform(p3B);
							m4View.transform(p3C);
							m4View.transform(p3D);
							drawable.setColor(settings.cSelected);
							drawable.drawCurveSegment(p3A,p3B,p3C,p3D);
						}
					}
				}
				
				if (viewDefinition.renderPatches()) {
					if (bZBuffer) {
						renderPatches(iMode);
					} else {
						updateImage();
						return;
					}
				}
				//paintTest();
				if (tool != null) tool.paint(this, drawable);
				drawActiveViewportMarker(MainFrame.getInstance().getJPatchScreen().getActiveViewport() == this);
				paintAxis();
				drawOrigin();
				drawDisplayName();
				
				Graphics2D g2 = (Graphics2D)drawable.getGraphics();
				
				/*
				 * output milliseconds
				 */
				g2.setColor(settings.cText);
				g2.setFont(font);
				g2.drawString("" + (System.currentTimeMillis() - start) +"ms", 7,30);
				
				/*
				 * copy image
				 */
				getGraphics().drawImage(image, 0, 0, this);
				
			}
		}
	}
}

