package net.sf.jpatch.boundary;

import java.awt.*;
import java.awt.geom.*;
import java.awt.image.*;
import javax.vecmath.*;
import net.sf.jpatch.entity.*;
import net.sf.jpatch.renderer.*;

public final class JPatchDrawableZBuffer extends ZBufferRenderer
implements JPatchDrawable {

	private static final int GHOST = JPatchSettings.getInstance().iGhost;
	
	ControlPoint cp;
	Point3f p3A = new Point3f();
	Point3f p3B = new Point3f();
	Point3f p3C = new Point3f();
	Point3f p3D = new Point3f();
		
	public JPatchDrawableZBuffer(BufferedImage image, Lighting lighting) {
		super(lighting);
		setImage(image);
		renderToImage();
	}

	/*
	public void drawTriangleArray3D(Point3f[] ap3, Color[] ac) {
		int c = 0;
		for (int p = 0; p < ap3.length;) {
			if (Line2D.relativeCCW(ap3[p].x, ap3[p].y, ap3[p + 1].x, ap3[p + 1].y, ap3[p + 2].x, ap3[p + 2].y) < 0) {
				draw3DTriangleFlat(ap3[p++], ap3[p++], ap3[p++], ac[c].getRGB());
			}
			c++;
		}
	}
	*/

	public final void drawSimpleShape(SimpleShape shape, Matrix4f matrix) {
		Point3f[] ap3Points = shape.getPoints();
		Vector3f[] av3Normals = shape.getNormals();
		int[] aiTriangles = shape.getTriangles();
		int[] aiLines = shape.getLines();
		
		MaterialProperties mp = shape.getMaterialProperties();
		
		if (matrix != null) {
			for (int i = 0; i < ap3Points.length; i++) {
				matrix.transform(ap3Points[i]);
			}
			for (int i = 0; i < av3Normals.length; i++) {
				matrix.transform(av3Normals[i]);
				av3Normals[i].normalize();
			}
		}
		
		/*
		setColor(new Color(mp.red, mp.green, mp.blue));
		for (int line = 0; line < aiLines.length;) {
			drawLine((int)ap3Points[aiLines[line]].x, (int)ap3Points[aiLines[line++]].y, (int)ap3Points[aiLines[line]].x, (int)ap3Points[aiLines[line++]].y);
		}
		*/
		int p = 0;
		for (int triangle = 0; triangle < av3Normals.length; triangle++) {
			if (av3Normals[triangle].z < 0) {
				int color = shade(null, av3Normals[triangle], mp);
				draw3DTriangleFlatGhost(ap3Points[aiTriangles[p++]], ap3Points[aiTriangles[p++]], ap3Points[aiTriangles[p++]], color, GHOST);
			} else {
				p += 3;
			}	
		}
	}
	
	//public final void drawJPatchCurve3D(Curve curve, Matrix4f matrix) {		
	//	cp = curve.getStart();
	//	p3A.set(cp.getPosition());
	//	p3B.set(cp.getOutTangent());
	//	matrix.transform(p3A);
	//	matrix.transform(p3B);
	//	cp = cp.getNext();
	//	while (cp != null) {
	//		p3C.set(cp.getInTangent());
	//		p3D.set(cp.getPosition());
	//		matrix.transform(p3C);
	//		matrix.transform(p3D);
	//		drawCurveSegment3D(p3A,p3B,p3C,p3D);
	//		ControlPoint cpNext = cp.getNextCheckLoop();
	//		if (cpNext != null) { 
	//			p3B.set(cp.getOutTangent());
	//			matrix.transform(p3B);
	//			p3A.set(p3D);
	//		}
	//		cp = cpNext;
	//	}
	//}
	
	public final void drawJPatchCurve3D(Curve curve, Matrix4f matrix) {		
		cp = curve.getStart();
		
		/* go to first not hidden point */
		while (cp != null && cp.isHidden()) {
			cp = cp.getNextCheckNextLoop();
		}
		
		if (cp != null && cp.getNext() != null) {
			p3A.set(cp.getPosition());
			p3B.set(cp.getOutTangent());
			matrix.transform(p3A);
			matrix.transform(p3B);
			cp = cp.getNext();
			loop:
			while (cp != null) {
				if (cp.isHidden()) {
						while (cp != null && cp.isHidden()) {
							cp = cp.getNextCheckLoop();
						}
						if (cp != null) {
							p3D.set(cp.getPosition());
							matrix.transform(p3D);
							//generalPath.moveTo(p3D.x, p3D.y);
						} else {
							break loop;
						}
					} else {
						p3C.set(cp.getInTangent());
						p3D.set(cp.getPosition());
						matrix.transform(p3C);
						matrix.transform(p3D);
						drawCurveSegment3D(p3A,p3B,p3C,p3D);
					}
				ControlPoint cpNext = cp.getNextCheckLoop();
				if (cpNext != null) { 
					p3B.set(cp.getOutTangent());
					matrix.transform(p3B);
					p3A.set(p3D);
				}
				cp = cpNext;
			}
		}
	}
	
}

