package net.sf.jpatch.boundary;

import java.util.*;
import java.lang.reflect.*;
import java.awt.*;
import java.awt.event.*;
import javax.vecmath.*;
import javax.swing.*;
import javax.swing.border.*;
import net.sf.jpatch.entity.*;
import net.sf.jpatch.boundary.mouse.*;
import net.sf.jpatch.boundary.tools.*;
import net.sf.jpatch.auxilary.*;

public final class JPatchScreen extends JPanel {
	public static final int SINGLE = 1;
	public static final int HORIZONTAL_SPLIT = 2;
	public static final int VERTICAL_SPLIT = 3;
	public static final int QUAD = 4;
	
	public static final int LIGHT_OFF = 0;
	public static final int LIGHT_SIMPLE = 1;
	public static final int LIGHT_HEAD = 2;
	public static final int LIGHT_THREE_POINT = 3;
	
	public static final int NUMBER_OF_VIEWPORTS = 4;
	
	private Viewport activeViewport;
	
	private JPatchCanvas[] aComponent = new JPatchCanvas[NUMBER_OF_VIEWPORTS];
	private ViewDefinition[] aViewDef;
	
	private boolean bSnapToGrid = JPatchSettings.getInstance().bGridSnap;
	private float fGridSpacing = JPatchSettings.getInstance().fGridSpacing;
	
	private Model model;
	private int iMode;
	//protected ControlPoint cpCursor = new ControlPoint(0,0,0);
	
	private boolean bSynchronized = JPatchSettings.getInstance().bSyncWindows;
	private int iLightMode = JPatchSettings.getInstance().iLightingMode;
	private boolean bStickyLight = JPatchSettings.getInstance().bStickyLight;
	private JPatchTool tool;
	
	private int[] aiView = new int[] {
			ViewDefinition.FRONT,
			ViewDefinition.LEFT,
			ViewDefinition.TOP,
			ViewDefinition.BIRDS_EYE
	};
	
	private PopupMouseListener popupMouseListener = new PopupMouseListener(MouseEvent.BUTTON3);
	private MoveZoomRotateMouseAdapter moveZoomRotateMouseAdapter = new MoveZoomRotateMouseAdapter();
	
	private boolean bPopupEnabled;
	private ActiveViewportMouseAdapter activeViewportMouseAdapter = new ActiveViewportMouseAdapter();
	private boolean bShowTangents = false;
	private TangentTool tangentTool = new TangentTool();
	
	public JPatchScreen(Model model,int mode,ViewDefinition[] viewDefinitions) {
		this.model = model;
		aViewDef = viewDefinitions;
		for (int i = 0; i < NUMBER_OF_VIEWPORTS; i++) {
			aComponent[i] = new JPatchCanvas(model, aViewDef[i]);
			//aComponent[i] = new JPatchCanvas(model,aViewDef[i]);
			add(aComponent[i]);
			aComponent[i].setFocusable(false);
		}
		setFocusable(false);
		setMode(mode);
		setLightingMode(iLightMode);
		enablePopupMenu(true);
		activeViewport = aComponent[0];
		snapToGrid(bSnapToGrid);
	}
	
	
				
	public boolean isSynchronized() {
		return bSynchronized;
	}
	
	public void synchronize(boolean sync) {
		bSynchronized = sync;
	}
	
	public boolean showTangents() {
		return bShowTangents;
	}
	
	public void showTangents(boolean enable) {
		bShowTangents = enable;
		for (int i = 0; i < NUMBER_OF_VIEWPORTS; i++) {
			if (enable && !bShowTangents) {
				aComponent[i].addMouseListener(tangentTool);
			} else {
				aComponent[i].removeMouseListener(tangentTool);
			}
		}
	}
	
	public TangentTool getTangentTool() {
		return tangentTool;
	}
	
	public void setLightingMode(int mode) {
		iLightMode = mode;
		for (int i = 0; i < NUMBER_OF_VIEWPORTS; i++) {
			switch(iLightMode) {
				case LIGHT_OFF:
					aComponent[i].flatShade(true);
					aComponent[i].setLighting(Lighting.createSimpleLight());
					break;
				case LIGHT_SIMPLE:
					aComponent[i].flatShade(false);
					aComponent[i].setLighting(Lighting.createSimpleLight());
					break;
				case LIGHT_HEAD:
					aComponent[i].flatShade(false);
					aComponent[i].setLighting(Lighting.createHeadLight());
					break;
				case LIGHT_THREE_POINT:
					aComponent[i].flatShade(false);
					aComponent[i].setLighting(Lighting.createThreePointLight());
					break;
			}
		}
		update_all();
	}
	
	public int getLightingMode() {
		return iLightMode;
	}
	
	public void setActiveViewport(Viewport viewport) {
		if (viewport != activeViewport) {
			activeViewport = viewport;
			//for (int i = 0; i < NUMBER_OF_VIEWPORTS; i++) {
			//	aComponent[i].drawActiveViewportMarker(aComponent[i] == viewport);
			//}
			update_all();
		}
	}
	
	public Viewport getActiveViewport() {
		return activeViewport;
	}
	
	public void setStickyLight(boolean sticky) {
		bStickyLight = sticky;
		for (int i = 0; i < NUMBER_OF_VIEWPORTS; i++) {
			aComponent[i].getLighting().setStickyLight(sticky);
		}
		update_all();
	}
	
	public boolean snapToGrid() {
		return bSnapToGrid;
	}
	
	public void snapToGrid(boolean enable) {
		bSnapToGrid = enable;
		for (int i = 0; i < NUMBER_OF_VIEWPORTS; i++) {
			aComponent[i].getGrid().snap(enable);
		}
		update_all();
		JPatchSettings.getInstance().bGridSnap = enable;
	}
	
	public float getGridSpacing() {
		return fGridSpacing;
	}
	
	public void setGridSpacing(float gridSpacing) {
		fGridSpacing = gridSpacing;
		for (int i = 0; i < NUMBER_OF_VIEWPORTS; i++) {
			aComponent[i].getGrid().setSpacing(gridSpacing);
		}
		update_all();
	}
	
	public boolean isStickyLight() {
		return bStickyLight;
	}
	
	public void single_update(Component component) {
		if (bSynchronized) {
			update_all();
		} else {
			((JPatchCanvas) component).render();
		}
	}
	
	public void full_update() {
		//System.out.println("full_update()");
		if (!bSynchronized) {
			update_all();
		}
	}
	
	public void update_all() {
		switch(iMode) {
			case SINGLE:
				aComponent[0].render();
				break;
			case HORIZONTAL_SPLIT:
				aComponent[0].render();
				aComponent[1].render();
				break;
			case VERTICAL_SPLIT:
				aComponent[0].render();
				aComponent[2].render();
				break;
			case QUAD:
				aComponent[0].render();
				aComponent[1].render();
				aComponent[2].render();
				aComponent[3].render();
				break;
		}
	}
	
	public void prepareBackground(Component component) {
		setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
		if (bSynchronized) {
			for (int i = 0; i < NUMBER_OF_VIEWPORTS; i++) {
				if (aComponent[i].isVisible()) {
					((JPatchCanvas)aComponent[i]).prepareBackground();
				}
			}
		} else {
			((JPatchCanvas)component).prepareBackground();
		}
		setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
	}
	
	public void clearBackground() {
		for (int i = 0; i < NUMBER_OF_VIEWPORTS; i++) {
			//if (aComponent[i].isVisible()) {
				((JPatchCanvas)aComponent[i]).clearBackground();
			//}
		}
	}
	
	//public void set3DCursor(Point3f cursor) {
	//	cpCursor.setPosition(cursor);
	//}
	//public void set3DCursor(float x, float y, float z) {
	//	cpCursor.setPosition(x,y,z);
	//}
	//public ControlPoint get3DCursor() {
	//	return cpCursor;
	//}
	
	public void addMouseListeners(MouseListener mouseAdapter) {
		for (int i = 0; i < NUMBER_OF_VIEWPORTS; i++) {
			aComponent[i].addMouseListener(mouseAdapter);
		}
	}
	
	public void setTool(JPatchTool tool) {
		this.tool = tool;
		removeAllMouseListeners();
		for (int i = 0; i < NUMBER_OF_VIEWPORTS; i++) {
			((JPatchCanvas)aComponent[i]).setTool(tool);
		}
		update_all();
	}
	
	public JPatchTool getTool() {
		return tool;
	}
	
	public void removeAllMouseListeners() {
		for (int i = 0; i < NUMBER_OF_VIEWPORTS; i++) {
			MouseListener[] aMouseListener = aComponent[i].getMouseListeners();
			MouseMotionListener[] aMouseMotionListener = aComponent[i].getMouseMotionListeners();
			MouseWheelListener[] aMouseWheelListener = aComponent[i].getMouseWheelListeners();
			for (int m = 0; m < aMouseListener.length; m++) {
				aComponent[i].removeMouseListener(aMouseListener[m]);
			}
			for (int m = 0; m < aMouseMotionListener.length; m++) {
				aComponent[i].removeMouseMotionListener(aMouseMotionListener[m]);
			}
			for (int m = 0; m < aMouseWheelListener.length; m++) {
				aComponent[i].removeMouseWheelListener(aMouseWheelListener[m]);
			}
		}
		enablePopupMenu(false);
		enablePopupMenu(true);
		addMMBListener();
	}
	/*
	public void rerender() {
		switch(mode) {
			case SINGLE:
				aComponent[0].render();
			break;
			case HORIZONTAL_SPLIT:
				aComponent[0].render();
				aComponent[1].render();
			break;
			case VERTICAL_SPLIT:
				aComponent[0].render();
				aComponent[2].render();
			break;
			case QUAD:
				aComponent[0].render();
				aComponent[1].render();
				aComponent[2].render();
				aComponent[3].render();
			break;
		}
	}
	*/
	public void enablePopupMenu(boolean enable) {
		//System.out.println("popup: " + bPopupEnabled + " " + enable);
		if (enable != bPopupEnabled) {
			for (int i = 0; i < NUMBER_OF_VIEWPORTS; i++) {
				if (enable) {
					aComponent[i].addMouseListener(popupMouseListener);
				} else {
					aComponent[i].removeMouseListener(popupMouseListener);
				}
			}
			bPopupEnabled = enable;
		}
	}
	
	public void addMMBListener() {
		for (int i = 0; i < NUMBER_OF_VIEWPORTS; i++) {
			aComponent[i].addMouseListener(moveZoomRotateMouseAdapter);
			aComponent[i].addMouseWheelListener(moveZoomRotateMouseAdapter);
			aComponent[i].addMouseListener(activeViewportMouseAdapter);
			aComponent[i].addMouseWheelListener(activeViewportMouseAdapter);
			if (bShowTangents) {
				aComponent[i].addMouseListener(tangentTool);
			}
		}
	}
	
	public int getMode() {
		return iMode;
	}
	
	public void setMode(int mode) {
		iMode = mode;
		int h = getHeight();
		int w = getWidth();
		int h2 = h/2;
		int w2 = w/2;
		aComponent[0].setVisible(false);
		aComponent[1].setVisible(false);
		aComponent[2].setVisible(false);
		aComponent[3].setVisible(false);
		aComponent[0].clearImage();
		aComponent[1].clearImage();
		aComponent[2].clearImage();
		aComponent[3].clearImage();
		//update_all();
		if (h > 0 && w > 0) {
			switch(mode) {
				case SINGLE:
					activeViewport = aComponent[0];
					aComponent[0].setBounds(0,0,w,h);
					aComponent[0].setVisible(true);
					//aComponent[1].setVisible(false);
					//aComponent[2].setVisible(false);
					//aComponent[3].setVisible(false);
				break;
				case HORIZONTAL_SPLIT:
					if (activeViewport != aComponent[0] && activeViewport != aComponent[1]) {
						activeViewport = aComponent[0];
					}
					aComponent[0].setBounds(0,0,w2 - 1,h);
					aComponent[0].setVisible(true);
					aComponent[1].setBounds(w2 + 1,0,w2 - 1,h);
					aComponent[1].setVisible(true);
					//aComponent[2].setVisible(false);
					//aComponent[3].setVisible(false);
				break;
				case VERTICAL_SPLIT:
					if (activeViewport != aComponent[0] && activeViewport != aComponent[2]) {
						activeViewport = aComponent[0];
					}
					aComponent[0].setBounds(0,0,w,h2 - 1);
					aComponent[0].setVisible(true);
					//aComponent[1].setVisible(false);
					aComponent[2].setBounds(0,h2 + 1,w,h2 - 1);
										aComponent[2].setVisible(true);

					//aComponent[3].setVisible(false);
				break;
				case QUAD:
					aComponent[0].setBounds(0,0,w2 - 1,h2 - 1);
					aComponent[0].setVisible(true);
					aComponent[1].setBounds(w2 + 1,0,w2 - 1,h2 - 1);
					aComponent[1].setVisible(true);
					aComponent[2].setBounds(0,h2 + 1,w2 - 1,h2 - 1);
					aComponent[2].setVisible(true);
					aComponent[3].setBounds(w2 + 1,h2 + 1,w2 - 1,h2 - 1);
					aComponent[3].setVisible(true);

				break;
			}
		}
		
		JPatchSettings.getInstance().iScreenMode = iMode;
		//JPatchSettings.getInstance().saveSettings();
		
		//update_all();
		/*
		iMode = mode;
		removeAll();
		switch(mode) {
			case SINGLE:
				setLayout(new GridLayout(1,1));
				add(aComponent[0], BorderLayout.CENTER);
				break;
			case HORIZONTAL_SPLIT:
				setLayout(new GridLayout(1,2));
				add(aComponent[0]);
				add(aComponent[1]);
				break;
			case VERTICAL_SPLIT:
				setLayout(new GridLayout(1,2));
				add(aComponent[0]);
				add(aComponent[2]);
				break;
			case QUAD:
				setLayout(new GridLayout(2,2));
				add(aComponent[0]);
				add(aComponent[1]);
				add(aComponent[2]);
				add(aComponent[3]);
				break;
		}
		validate();
		*/
	}
	
	public void doLayout() {
		setMode(iMode);
	}
	
	public void resetMode(int mode) {
		setMode(mode);
	}
}
