package net.sf.jpatch.boundary;

import java.util.*;
import java.util.prefs.*;
import java.awt.*;

public class JPatchSettings {
	
	public boolean bFirstStart;
	
	public Color cBackground;
	public Color cCurve;
	public Color cPoint;
	public Color cHeadPoint;
	public Color cPointSelected;
	public Color cSelection;
	public Color cText;
	public Color cTangent;
	public Color cOrigin;
	public Color cGrid;
	public Color cGridMin;
	public Color cX;
	public Color cY;
	public Color cZ;
	public Color cSelected;
	public Color cHot;
	public Color cGrey;
	
	public int iGhost = 0xA0;
	
	public int iBackground;
	public int iCurve;
	public int iPoint;
	public int iPointSelected;
	public int iGrid;
	public int iGridMin;
	public int iSelection;
	public int iX;
	public int iY;
	public int iZ;
	
	public int iScreenMode;
	public boolean bSyncWindows;
	
	public int iPatchSubdivisions;
	public int iCurveSubdivisions;

	public int iScreenX;
	public int iScreenY;
	public int iScreenWidth;
	public int iScreenHeight;
	
	public int iLightingMode;
	public boolean bStickyLight;
	
	public boolean bGridSnap;
	public float fGridSpacing;
	
	public String strJPatchPath;
	public String strAMPath;
	public String strSPatchPath;
	public String strPovrayPath;
	public String strRendermanPath;
	public String strRotoscopePath;
	
	public String strJPatchFile = "";
	
	public String[] astrKeyMap = new String[] {
		"Single View			Single view			1",
		"Horizontal Split View		Horizontal split view		2",
		"Vertical Split View		Vertical split view		3",
		"Quad View			Quad view			4",
		"Rotate View			Rotate view			SHIFT-R",
		"Move View			Move view			SHIFT-M",
		"Zoom View			Zoom view			SHIFT-Z",
		"Undo				Undo				CTRL-Z",
		"Redo				Redo				CTRL-Y",
		"Lock X				Lock/unlock X			X",
		"Lock Y				Lock/unlock Y			Y",
		"Lock Z				Lock/unlock Z			Z",
		"Default Tool			Default tool			ESC",
		"Add Point			Add point			A",
		"Add Multiple Points		Add multiple points		SHIFT-A",
		"Magnet Tool			Magnet tool			M",
		"Rotate Tool			Rotate tool			R",
		"Remove Points			Remove points			DEL",
		"Delete Points			Delete points			BACKSPACE",
		"Compute Patches		Compute patches			F5",
		"Clone				Copy				CTRL-C",
		"Extrude			Extrude				E",
		"Lathe				Lathe				L",
		"Extend Selection		Extend selection		NUM+",
		"Next Curve			Next curve			TAB",
		"Insert Point			Insert point			INS",
		"Bottom View			Bottom view			NUM0",
		"Front View			Front view			NUM2",
		"Right View			Right view			NUM4",
		"Top View			Top view			NUM5",
		"Left View			Left view			NUM6",
		"Bird's Eye View		Bird's eye view			NUM7",
		"Rear View			Rear view			NUM8",
		"Grid				Snap to grid			G",
		"Hide				Hide/show unselected points	H"
	};
	
	private static JPatchSettings INSTANCE;
	private Preferences userPrefs = Preferences.userRoot().node("/net/sf/jpatch/Preferences");
	private Map mapDefaults = new HashMap();
	
	/**
	 * constructor
	 */
	public JPatchSettings() {
		mapDefaults.put("firstStart", new Boolean(true));
		
		mapDefaults.put("backgroundColor",new Color(0x30,0x40,0x50));
		mapDefaults.put("curveColor",new Color(255,255,255));
		mapDefaults.put("pointColor",new Color(255,255,0));
		mapDefaults.put("headPointColor",new Color(255,0,0));
		mapDefaults.put("tangentColor",new Color(255,255,0));
		mapDefaults.put("selectedPointColor",new Color(0,255,0));
		mapDefaults.put("selectionColor",new Color(255,255,0));
		mapDefaults.put("textColor",new Color(0x80,0x90,0xA0));
		mapDefaults.put("originColor",new Color(0xFF,0xFF,0xFF));
		mapDefaults.put("gridColor",new Color(0x10,0x20,0x30));
		mapDefaults.put("minorGridColor",new Color(0x20,0x30,0x40));
		mapDefaults.put("xColor",new Color(255,64,0));
		mapDefaults.put("yColor",new Color(0,255,0));
		mapDefaults.put("zColor",new Color(128,128,255));
		mapDefaults.put("selectedColor",new Color(0,255,0));
		mapDefaults.put("hotColor",new Color(0,255,255));
		mapDefaults.put("greyColor",new Color(0x50,0x60,0x70));
		
		mapDefaults.put("screenMode", new Integer(JPatchScreen.SINGLE));
		mapDefaults.put("syncWindows", new Boolean(false));
		
		mapDefaults.put("patchSubdivisions", new Integer(3));
		mapDefaults.put("curveSubdivisions", new Integer(5));
		
		mapDefaults.put("screenX", new Integer(0));
		mapDefaults.put("screenY", new Integer(0));
		mapDefaults.put("screenWidth", new Integer(1000));
		mapDefaults.put("screenHeight", new Integer(700));
		
		mapDefaults.put("lightingMode", new Integer(JPatchScreen.LIGHT_THREE_POINT));
		mapDefaults.put("stickyLight", new Boolean(true));
		
		mapDefaults.put("gridSnap", new Boolean(false));
		mapDefaults.put("gridSpacing", new Float(1));
		
		mapDefaults.put("jpatchPath", "");
		mapDefaults.put("spatchPath", "");
		mapDefaults.put("amPath", "");
		mapDefaults.put("povrayPath", "");
		mapDefaults.put("rendermanPath", "");
		mapDefaults.put("rotoscopePath", "");
		
		loadSettings();
		saveSettings();
	}
	
	/**
	 * returns the instance (singleton pattern)
	 */
	public static JPatchSettings getInstance() {
		if (INSTANCE == null) {
			INSTANCE = new JPatchSettings();
		}
		return INSTANCE;
	}
	
	/**
	 * load all setings from backings store
	 */
	public void loadSettings() {
		bFirstStart = getBoolean("firstStart");
		
		cBackground = getColor("backgroundColor");
		cCurve = getColor("curveColor");
		cPoint = getColor("pointColor");
		cTangent = getColor("tangentColor");
		cHeadPoint = getColor("headPointColor");
		cPointSelected = getColor("selectedPointColor");
		cSelection = getColor("selectionColor");
		cText = getColor("textColor");
		cOrigin = getColor("originColor");
		cGrid =	getColor("gridColor");
		cGridMin = getColor("minorGridColor");
		cX = getColor("xColor");
		cY = getColor("yColor");
		cZ = getColor("zColor");
		cSelected = getColor("selectedColor");
		cHot = getColor("hotColor");
		cGrey = getColor("greyColor");
		
		iBackground = cBackground.getRGB();
		iCurve = cCurve.getRGB();
		iPoint = cPoint.getRGB();
		iPointSelected = cPointSelected.getRGB();
		iGrid = cGrid.getRGB();
		iGridMin = cGridMin.getRGB();
		iSelection = cSelection.getRGB();
		iX = cX.getRGB();
		iY = cY.getRGB();
		iZ = cZ.getRGB();
		
		iScreenMode = getInt("screenMode");
		bSyncWindows = getBoolean("syncWindows");
		
		iPatchSubdivisions = getInt("patchSubdivisions");
		iCurveSubdivisions = getInt("curveSubdivisions");
		
		iScreenX = getInt("screenX");
		iScreenY = getInt("screenY");
		iScreenWidth = getInt("screenWidth");
		iScreenHeight = getInt("screenHeight");
		
		iLightingMode = getInt("lightingMode");
		bStickyLight = getBoolean("stickyLight");
		
		bGridSnap = getBoolean("gridSnap");
		fGridSpacing = getFloat("gridSpacing");
		
		strJPatchPath = getString("jpatchPath");
		strSPatchPath = getString("spatchPath");
		strAMPath = getString("amPath");
		strPovrayPath = getString("povrayPath");
		strRendermanPath = getString("rendermanPath");
		strRotoscopePath = getString("rotoscopePath");
	}
	
	/**
	 * save all settings to backing store
	 */
	public void saveSettings() {
		putBoolean("firstStart", bFirstStart);
		
		putColor("backgroundColor", cBackground);
		putColor("curveColor", cCurve);
		putColor("pointColor", cPoint);
		putColor("headPointColor", cHeadPoint);
		putColor("selectedPointColor", cPointSelected);
		putColor("selectionColor", cSelection);
		putColor("textColor", cText);
		putColor("originColor", cOrigin);
		putColor("gridColor", cGrid);
		putColor("minorGridColor", cGridMin);
		putColor("xColor", cX);
		putColor("yColor", cY);
		putColor("zColor", cZ);
		putColor("selectedColor", cSelected);
		putColor("hotColor", cHot);
		putColor("tangentColor", cTangent);
		
		putInt("screenMode", iScreenMode);
		putBoolean("syncWindows", bSyncWindows);
		
		putInt("curveSubdivisions", iCurveSubdivisions);
		putInt("patchSubdivisions", iPatchSubdivisions);

		putInt("screenX", iScreenX);
		putInt("screenY", iScreenY);
		putInt("screenWidth", iScreenWidth);
		putInt("screenHeight", iScreenHeight);
		
		putInt("lightingMode", iLightingMode);
		putBoolean("stickyLight", bStickyLight);
		putBoolean("gridSnap", bGridSnap);
		putFloat("gridSpacing", fGridSpacing);
		
		putString("jpatchPath", strJPatchPath);
		putString("spatchPath", strSPatchPath);
		putString("amPath", strAMPath);
		putString("povrayPath", strPovrayPath);
		putString("rendermanPath", strRendermanPath);
		putString("rotoscopePath", strRotoscopePath);
		
		try {
			userPrefs.flush();
		} catch (BackingStoreException exception) {
		}
	}
	
	private void putInt(String key, int i) {
		userPrefs.putInt(key, i);
	}
	
	private int getInt(String key) {
		return userPrefs.getInt(key,((Integer) mapDefaults.get(key)).intValue());
	}
	
	private void putFloat(String key, float f) {
		userPrefs.putFloat(key, f);
	}
	
	private float getFloat(String key) {
		return userPrefs.getFloat(key,((Float) mapDefaults.get(key)).floatValue());
	}
	
	private void putBoolean(String key, boolean b) {
		userPrefs.putBoolean(key, b);
	}
	
	private boolean getBoolean(String key) {
		return userPrefs.getBoolean(key,((Boolean) mapDefaults.get(key)).booleanValue());
	}
	
	private void putString(String key, String value) {
		userPrefs.put(key, value);
	}
	
	private String getString(String key) {
		return userPrefs.get(key, (String) mapDefaults.get(key));
	}
	
	
		
	/**
	 * put an awt.Color to preferences node
	 */
	private void putColor(String key, Color color) {
		userPrefs.putInt(key, color.getRGB());
	}

	/**
	 * get an awt.Color from preferences node
	 */
	private Color getColor(String key) {
		//Color c = (Color) mapDefaults.get(key);
		//System.out.println(key + " " + c);
		//return new Color(userPrefs.getInt(key,((Color) mapDefaults.get(key)).getRGB()));
		return (Color) mapDefaults.get(key);
	}
}
