package net.sf.jpatch.boundary;

import javax.swing.*;
import java.awt.*;
import net.sf.jpatch.boundary.action.*;

public final class MainMenu extends JMenuBar {
	private JMenu menuFile;
	private JMenu menuImport;
	private JMenu menuExport;
	private JMenu menuTest;
	private JMenu menuHelp;
	private JMenu menuOptions;
	private JMenu menuLight;
	
	private JCheckBoxMenuItem miLightOff;
	private JCheckBoxMenuItem miLightSimple;
	private JCheckBoxMenuItem miLightHead;
	private JCheckBoxMenuItem miLightThreePoint;
	private JCheckBoxMenuItem miLightSticky;
	
	//private Action actionComputePatches = new ComputePatchesAction();
	//private Action actionExtendCurve = new NextCurveAction();
	//private Action actionNextCurve = new NextCurveAction();
	
	public MainMenu() {

		
		menuFile = new JMenu("File");
		menuImport = new JMenu("Import");
		menuExport = new JMenu("Export");
		menuTest = new JMenu("Test");
		menuHelp = new JMenu("Help");
		menuOptions = new JMenu("Options");	
		menuLight = new JMenu("Light");
		
		miLightOff = new JCheckBoxMenuItem(new LightingOffAction(MainFrame.getInstance().getJPatchScreen()));
		miLightSimple = new JCheckBoxMenuItem(new LightingSimpleAction(MainFrame.getInstance().getJPatchScreen()));
		miLightHead = new JCheckBoxMenuItem(new LightingHeadAction(MainFrame.getInstance().getJPatchScreen()));
		miLightThreePoint = new JCheckBoxMenuItem(new LightingThreePointAction(MainFrame.getInstance().getJPatchScreen()));
		miLightSticky = new JCheckBoxMenuItem(new LightingStickyAction(MainFrame.getInstance().getJPatchScreen()));
		
		ButtonGroup bgLight = new ButtonGroup();
		bgLight.add(miLightOff);
		bgLight.add(miLightSimple);
		bgLight.add(miLightHead);
		bgLight.add(miLightThreePoint);
		
		miLightOff.setState(MainFrame.getInstance().getJPatchScreen().getLightingMode() == JPatchScreen.LIGHT_OFF);
		miLightSimple.setState(MainFrame.getInstance().getJPatchScreen().getLightingMode() == JPatchScreen.LIGHT_SIMPLE);
		miLightHead.setState(MainFrame.getInstance().getJPatchScreen().getLightingMode() == JPatchScreen.LIGHT_HEAD);
		miLightThreePoint.setState(MainFrame.getInstance().getJPatchScreen().getLightingMode() == JPatchScreen.LIGHT_THREE_POINT);
		miLightSticky.setState(MainFrame.getInstance().getJPatchScreen().isStickyLight());
		
		menuLight.add(miLightOff);
		menuLight.add(miLightSimple);
		menuLight.add(miLightHead);
		menuLight.add(miLightThreePoint);
		menuLight.addSeparator();
		menuLight.add(miLightSticky);
		
		menuImport.add(new ImportSPatchAction());
		menuImport.add(new ImportAnimationMasterAction());
		//menuImport.add(new ImportJPatchAction());
		
		JMenuItem mi;
		mi = new JMenuItem(new NewAction());
		mi.setIcon(null);
		mi.setText("New");
		menuFile.add(mi);
		mi = new JMenuItem(new ImportJPatchAction());
		mi.setIcon(null);
		mi.setText("Open");
		menuFile.add(mi);
		mi = new JMenuItem(new SaveAsAction(false));
		mi.setIcon(null);
		mi.setText("Save");
		menuFile.add(mi);
		mi = new JMenuItem(new SaveAsAction(true));
		mi.setIcon(null);
		mi.setText("Save As...");
		menuFile.add(mi);
		menuFile.addSeparator();
		menuFile.add(menuImport);
		
		menuTest.add(new DumpAction());
		menuTest.add(new DumpUndoStackAction());
		//menuTest.add(actionComputePatches);
		//menuTest.add(new TestMakeFivePointPatchAction());
		//menuTest.add(new ExtendSelectionAction());
		//menuTest.add(new NextCurveAction());
		//menuTest.add(new CloneAction());
		menuExport.add(new TestRenderPovRayAction());
		menuExport.add(new TestRenderRibAction());
		menuExport.add(new GenericOutputAction());
		//menuExport.add(new TestRenderTest2Action());
		//menuTest.add(new TestXMLoutAction());
		//menuTest.add(new TestSetQualityAction());
		
		menuFile.add(menuExport);
		menuFile.addSeparator();
		menuFile.add(new QuitAction());
		
		menuHelp.add(new AboutAction());
		
		menuOptions.add(new SetGridSpacingAction());
		menuOptions.add(new ZBufferQualityAction());
		menuOptions.add(menuLight);
		
		add(menuFile);
		add(menuOptions);
		add(menuTest);
		add(menuHelp);
		
		//MainFrame.getInstance().getKeyAdapter().addKey(KeyMapping.getKey("compute patches"), actionComputePatches);
		MainFrame.getInstance().getKeyAdapter().addKey(KeyMapping.getKey("next curve"), new NextCurveAction());
		MainFrame.getInstance().getKeyAdapter().addKey(KeyMapping.getKey("extend selection"), new ExtendSelectionAction());
		MainFrame.getInstance().getKeyAdapter().addKey(KeyMapping.getKey("insert point"), new InsertControlPointAction());
		
		MainFrame.getInstance().getKeyAdapter().addKey(KeyMapping.getKey("bottom view"), new ViewKeyAction(ViewDefinition.BOTTOM));
		MainFrame.getInstance().getKeyAdapter().addKey(KeyMapping.getKey("front view"), new ViewKeyAction(ViewDefinition.FRONT));
		MainFrame.getInstance().getKeyAdapter().addKey(KeyMapping.getKey("right view"), new ViewKeyAction(ViewDefinition.RIGHT));
		MainFrame.getInstance().getKeyAdapter().addKey(KeyMapping.getKey("top view"), new ViewKeyAction(ViewDefinition.TOP));
		MainFrame.getInstance().getKeyAdapter().addKey(KeyMapping.getKey("left view"), new ViewKeyAction(ViewDefinition.LEFT));
		MainFrame.getInstance().getKeyAdapter().addKey(KeyMapping.getKey("bird's eye view"), new ViewKeyAction(ViewDefinition.BIRDS_EYE));
		MainFrame.getInstance().getKeyAdapter().addKey(KeyMapping.getKey("rear view"), new ViewKeyAction(ViewDefinition.REAR));
		
		/*
		Font font = new Font("SansSerif",Font.PLAIN,12);
		setAllFonts(font,menuFile);
		setAllFonts(font,menuDisplay);
		setAllFonts(font,menuTest);
		*/
	}
	/*
	private void setAllFonts(Font font, JMenu menu) {
		MenuElement[] aMenuElement = menu.getSubElements();
		for (int e = 0; e < aMenuElement.length; e++) {
			try {
				if (aMenuElement[e].getClass() == Class.forName("JMenu")) {
					setAllFonts(font,(JMenu)aMenuElement[e]);
				} else {
					((JMenuItem)aMenuElement[e]).setFont(font);
				}
			} catch(Exception exception) {
			}
			menu.setFont(font);
		}
	}
	*/
}
		
		
