package net.sf.jpatch.boundary.dialog;

import java.awt.*;
import java.awt.event.*;
import javax.swing.event.*;
import javax.swing.*;
import net.sf.jpatch.entity.*;
import net.sf.jpatch.boundary.*;

public class BiasDialog extends JDialog 
implements ChangeListener, ItemListener, ActionListener {
	
	private String[] astrMode = { "peak","round (sPatch)","smooth (A:M)","free (G0)","default (G1)","equal (C1)" };
	private boolean[] abInCurvature = { true,true,true,true,true,true };
	private boolean[] abOutCurvature = { false,false,false,true,true,false };
	private boolean[] abInAlpha = { false,false,true,true,true,true };
	private boolean[] abOutAlpha = { false,false,false,true,false,false };
	private boolean[] abInGamma = { false,false,true,true,true,true };
	private boolean[] abOutGamma = { false,false,false,true,false,false };
	private ControlPoint cp = null;
	private JLabel label = new JLabel();
	private JPatchSlider sliderInCurvature;
	private JPatchSlider sliderOutCurvature;
	private JPatchSlider sliderInAlpha;
	private JPatchSlider sliderOutAlpha;
	private JPatchSlider sliderInGamma;
	private JPatchSlider sliderOutGamma;
	private JComboBox comboboxMode = new JComboBox(astrMode);
	private JButton buttonReset = new JButton("Reset");
	private int iMode = 4;
	
	public BiasDialog() {
		this(null);
	}
	
	public BiasDialog(ControlPoint cp) {
		super(MainFrame.getInstance(),"Tangent control");
		//setLocationRelativeTo(MainFrame.getInstance());
		this.cp = cp;
		setSize(260,200);
		reInit();
		setDefaultCloseOperation(DO_NOTHING_ON_CLOSE);
		JPatchSlider.setDimensions(90,100,50,20);
		sliderInCurvature = new JPatchSlider("in curvature",0,1,1f/3f);
		sliderOutCurvature = new JPatchSlider("out curvature",0,1,1f/3f);
		sliderInAlpha = new JPatchSlider("in alpha",-180,180,0);
		sliderOutAlpha = new JPatchSlider("out alpha",-180,180,0);
		sliderInGamma = new JPatchSlider("in gamma",-180,180,0);
		sliderOutGamma = new JPatchSlider("out gamma",-180,180,0);
		sliderInCurvature.addChangeListener(this);
		sliderOutCurvature.addChangeListener(this);
		sliderInAlpha.addChangeListener(this);
		sliderOutAlpha.addChangeListener(this);
		sliderInGamma.addChangeListener(this);
		sliderOutGamma.addChangeListener(this);
		comboboxMode.addItemListener(this);
		buttonReset.addActionListener(this);
		show();
	}
	
	public void setControlPoint (ControlPoint cp) {
		if (cp != this.cp) {
			this.cp = cp;
			reInit();
		}
	}
	
	public void reInit() {
		Container contentPane = getContentPane();
		Component[] aComponent = contentPane.getComponents();
		for (int c = 0; c < aComponent.length; c++) {
			contentPane.remove(aComponent[c]);
		}
		contentPane.setLayout(null);
		if (cp == null) {
			System.out.println("Y++");
			label.setText("select a controlpoint");
			contentPane.add(label);
			label.setBounds(0,0,260,20);
		} else {
			label.setText("ControlPoint@" + cp.hashCode());
			iMode = cp.getMode();
			comboboxMode.setSelectedIndex(iMode);
			sliderInCurvature.setValue(cp.getInCurvature());
			sliderOutCurvature.setValue(cp.getOutCurvature());
			sliderInAlpha.setValue(cp.getInAlpha());
			sliderOutAlpha.setValue(cp.getOutAlpha());
			sliderInGamma.setValue(cp.getInGamma());
			sliderOutGamma.setValue(cp.getOutGamma());
			addComponentBounds(label,0,0,260,20);
			addComponentBounds(sliderInCurvature,0,20,260,20);
			addComponentBounds(sliderOutCurvature,0,40,260,20);
			addComponentBounds(sliderInAlpha,0,60,260,20);
			addComponentBounds(sliderOutAlpha,0,80,260,20);
			addComponentBounds(sliderInGamma,0,100,260,20);
			addComponentBounds(sliderOutGamma,0,120,260,20);
			comboboxMode.setEditable(false);
			addComponentBounds(new JLabel("Mode:"),0,150,50,20);
			addComponentBounds(comboboxMode,40,150,140,20);
			addComponentBounds(buttonReset,180,150,70,20);
			
			sliderInCurvature.setEnabled(abInCurvature[iMode]);
			sliderOutCurvature.setEnabled(abOutCurvature[iMode]);
			sliderInAlpha.setEnabled(abInAlpha[iMode]);
			sliderOutAlpha.setEnabled(abOutAlpha[iMode]);
			sliderInGamma.setEnabled(abInGamma[iMode]);
			sliderOutGamma.setEnabled(abOutGamma[iMode]);
			
		}
		validate();
		repaint();
		MainFrame.getInstance().getJPatchScreen().repaint();
	}
	
	private void addComponentBounds(Component component,int x, int y, int w, int h) {
		getContentPane().add(component);
		component.setBounds(x,y,w,h);
	}
	
	public void stateChanged(ChangeEvent changeEvent) {
		if (changeEvent.getSource() == sliderInCurvature) {
			cp.setInCurvature(sliderInCurvature.getValue());
			if (!abOutCurvature[iMode]) {
				sliderOutCurvature.setValue(sliderInCurvature.getValue());
				cp.setOutCurvature(sliderInCurvature.getValue());
			}
		} else if (changeEvent.getSource() == sliderOutCurvature) {
			cp.setOutCurvature(sliderOutCurvature.getValue());
		} else if (changeEvent.getSource() == sliderInAlpha) {
			cp.setInAlpha(sliderInAlpha.getValue());
			if (!abOutAlpha[iMode]) {
				sliderOutAlpha.setValue(sliderInAlpha.getValue());
				cp.setOutAlpha(sliderInAlpha.getValue());
			}
		} else if (changeEvent.getSource() == sliderOutAlpha) {
			cp.setOutAlpha(sliderOutAlpha.getValue());
		} else if (changeEvent.getSource() == sliderInGamma) {
			cp.setInGamma(sliderInGamma.getValue());
			if (!abOutGamma[iMode]) {
				sliderOutGamma.setValue(sliderInGamma.getValue());
				cp.setOutGamma(sliderInGamma.getValue());
			}
		} else if (changeEvent.getSource() == sliderOutGamma) {
			cp.setOutGamma(sliderOutGamma.getValue());
		}
		MainFrame.getInstance().getJPatchScreen().repaint();
	}
	
	public void itemStateChanged(ItemEvent itemEvent) {
		if (itemEvent.getStateChange() == ItemEvent.SELECTED) {
			for (int i = 0; i < astrMode.length; i++) {
				if (astrMode[i].equals((String)itemEvent.getItem())) {
					cp.setMode((byte)i);
					setMode(i);
				}
			}
		}
	}
	
	public void actionPerformed(ActionEvent actionEvent) {
		if (actionEvent.getSource() == buttonReset) {
			reset();
		}
	}
	
	public void reset() {
		//iMode = 4:
		cp.setInCurvature(1f/3f);
		cp.setOutCurvature(1f/3f);
		cp.setInAlpha(0);
		cp.setOutAlpha(0);
		cp.setInGamma(0);
		cp.setOutGamma(0);
		reInit();
	}
	
	public void setMode (int mode) {
		iMode = mode;
		switch(mode) {
			case 0:
			cp.setOutCurvature(cp.getInCurvature());
			cp.setInAlpha(0);
			cp.setOutAlpha(0);
			cp.setInGamma(0);
			cp.setOutGamma(0);
			break;
			case 1:
			cp.setOutCurvature(cp.getInCurvature());
			cp.setInAlpha(0);
			cp.setOutAlpha(0);
			cp.setInGamma(0);
			cp.setOutGamma(0);
			break;
			case 2:
			cp.setOutCurvature(cp.getInCurvature());
			cp.setOutAlpha(cp.getInAlpha());
			cp.setOutGamma(cp.getInGamma());
			break;
			case 4:
			cp.setOutAlpha(cp.getInAlpha());
			cp.setOutGamma(cp.getInGamma());
			break;
			case 5:
			cp.setOutCurvature(cp.getInCurvature());
			cp.setOutAlpha(cp.getInAlpha());
			cp.setOutGamma(cp.getInGamma());
		}
		reInit();
		
	}
}

