package net.sf.jpatch.boundary.dialog;

import java.awt.*;
import java.awt.event.*;
import javax.swing.event.*;
import javax.swing.*;
import javax.swing.undo.*;
import javax.vecmath.*;
import net.sf.jpatch.entity.*;
import net.sf.jpatch.boundary.*;
import net.sf.jpatch.boundary.mouse.*;
import net.sf.jpatch.boundary.selection.*;
import net.sf.jpatch.control.edit.*;

public class RotateDialog extends JDialog 
implements ItemListener, ChangeListener, ActionListener {
	
	private String[] astrPivot = { "selection center","origin","3D cursor" };
	private String[] astrAxis = { "window Z","window X/Y","X","Y","Z" };
	//private String[] astrSnap = { "off","5","15","45" };
	
	public static final int CENTER = 0;
	public static final int ORIGIN = 1;
	public static final int CURSOR = 2;
	public static final int WINDOW_Z = 0;
	public static final int WINDOW_XY = 1;
	public static final int X = 2;
	public static final int Y = 3;
	public static final int Z = 4;
	
	public static final Vector3f[] av3Axis = new Vector3f[] {
		null,
		null,
		new Vector3f(1,0,0),
		new Vector3f(0,1,0),
		new Vector3f(0,0,1)
	};
	
	private static int iPivot = CENTER;
	private static int iAxis = WINDOW_Z;
	//private static int iSnap = 0;
	
	private RotateMouseAdapter rotateMouseAdapter;
	private CompoundEdit compoundEdit;
	
	private JComboBox comboboxPivot = new JComboBox(astrPivot);
	private JComboBox comboboxAxis = new JComboBox(astrAxis);
	//private JComboBox comboboxSnap = new JComboBox(astrSnap);
	private JButton buttonApply = new JButton("OK");
	private JButton buttonReset = new JButton("Cancel");
	
	private JPatchSlider rotateSlider;
	
	public RotateDialog(CompoundEdit compoundEdit) {
		super(MainFrame.getInstance(),"Rotate");
		//setLocationRelativeTo(MainFrame.getInstance());
		this.compoundEdit = compoundEdit;
		setSize(200,140);
		setDefaultCloseOperation(DO_NOTHING_ON_CLOSE);
		getContentPane().setLayout(null);
		addComponentBounds(new JLabel("pivot:"),0,0,50,20);
		addComponentBounds(comboboxPivot,50,0,122,20);
		addComponentBounds(new JLabel("axis:"),0,20,50,20);
		addComponentBounds(comboboxAxis,50,20,122,20);
		addComponentBounds(buttonReset,0,80,80,20);
		addComponentBounds(buttonApply,92,80,80,20);
		//addComponentBounds(new JLabel("snap:"),0,40,50,20);
		//addComponentBounds(comboboxSnap,50,40,170,20);
		JPatchSlider.setDimensions(50,72,50,20);
		rotateSlider = new JPatchSlider("rotate:",-180,180,0);
		addComponentBounds(rotateSlider,0,50,260,20);
		comboboxPivot.addItemListener(this);
		comboboxAxis.addItemListener(this);
		comboboxPivot.setSelectedIndex(iPivot);
		comboboxAxis.setSelectedIndex(iAxis);
		//comboboxSnap.addItemListener(this);
		rotateSlider.addChangeListener(this);
		rotateSlider.setEnabled(iAxis > 1);
		buttonReset.addActionListener(this);
		buttonApply.addActionListener(this);
		show();
	}
	
	public void setRotateMouseAdapter(RotateMouseAdapter rotateMouseAdapter) {
		this.rotateMouseAdapter = rotateMouseAdapter;
	}
	
	public void stateChanged(ChangeEvent changeEvent) {
		if (changeEvent.getSource() == rotateSlider) {
			rotateMouseAdapter.rotate(MainFrame.getInstance().getSelection(),new AxisAngle4f(av3Axis[iAxis],rotateSlider.getValue() / 180 * (float)Math.PI));
			MainFrame.getInstance().getJPatchScreen().repaint();
		}
	}
			
	private void addComponentBounds(Component component,int x, int y, int w, int h) {
		getContentPane().add(component);
		component.setBounds(x,y,w,h);
	}
	
	public void itemStateChanged(ItemEvent itemEvent) {
		if (itemEvent.getStateChange() == ItemEvent.SELECTED) {
			if (itemEvent.getSource() == comboboxPivot) {
				for (int i = 0; i < astrPivot.length; i++) {
					if (astrPivot[i].equals((String)itemEvent.getItem())) {
						iPivot = i;
					}
				}
			} else if (itemEvent.getSource() == comboboxAxis) {
				for (int i = 0; i < astrAxis.length; i++) {
					if (astrAxis[i].equals((String)itemEvent.getItem())) {
						iAxis = i;
					}
				}
				rotateSlider.setEnabled(iAxis > 1);
			}
			ControlPoint[] acp = ((PointSelection)MainFrame.getInstance().getSelection()).getControlPointArray();
			compoundEdit.addEdit(new MoveControlPointsEdit(MoveControlPointsEdit.ROTATE,acp));
		}
	}
	
	public void actionPerformed(ActionEvent actionEvent) {
		if (actionEvent.getSource() == buttonReset) {
			compoundEdit.end();
			if (compoundEdit.canUndo()) {
				compoundEdit.undo();
				MainFrame.getInstance().getJPatchScreen().repaint();
			}
			MainFrame.getInstance().getMeshToolBar().reset();
		} else if (actionEvent.getSource() == buttonApply) {
			ControlPoint[] acp = ((PointSelection)MainFrame.getInstance().getSelection()).getControlPointArray();
			compoundEdit.addEdit(new MoveControlPointsEdit(MoveControlPointsEdit.ROTATE,acp));
			compoundEdit.end();
			MainFrame.getInstance().getUndoManager().addEdit(compoundEdit);
			MainFrame.getInstance().getMeshToolBar().reset();
		}
	}
	
	public int getPivot() {
		return iPivot;
	}
	
	public int getAxis() {
		return iAxis;
	}
		
	public void dispose() {
		if (compoundEdit.isInProgress()) {
			ControlPoint[] acp = ((PointSelection)MainFrame.getInstance().getSelection()).getControlPointArray();
			compoundEdit.addEdit(new MoveControlPointsEdit(MoveControlPointsEdit.ROTATE,acp));
			compoundEdit.end();
			MainFrame.getInstance().getUndoManager().addEdit(compoundEdit);
		}
		super.dispose();
	}
}
