package net.sf.jpatch.boundary.mouse;

import java.awt.*;
import java.awt.image.*;
import java.awt.geom.*;
import java.awt.event.*;
import javax.swing.undo.*;
import javax.vecmath.*;
import net.sf.jpatch.entity.*;
import net.sf.jpatch.boundary.*;
import net.sf.jpatch.boundary.selection.*;
import net.sf.jpatch.boundary.dialog.*;
import net.sf.jpatch.control.edit.*;
import java.io.*;

public class RotateMouseAdapter extends JPatchMouseAdapter {
	
	private Component compSource;
	private RotateDialog rotateDialog;
	
	private int iMouseX;
	private int iMouseY;
	private Vector3f v3Axis = new Vector3f();
	private AxisAngle4f axisAngle = new AxisAngle4f();
	private AxisAngle4f axisAngle2 = new AxisAngle4f();
	private ControlPoint[] acp;
	private Point3f[] ap3;
	private Point3f p3Center;
	private Matrix4f m4InvView = new Matrix4f();
	
	private CompoundEdit compoundEdit;
	
	public RotateMouseAdapter(RotateDialog rotateDialog, CompoundEdit compoundEdit) {
		this.rotateDialog = rotateDialog;
		this.compoundEdit = compoundEdit;
		rotateDialog.setRotateMouseAdapter(this);
		Class classPointSelection = PointSelection.getPointSelectionClass();
		Selection selection = MainFrame.getInstance().getSelection();
		if (classPointSelection.isAssignableFrom(selection.getClass())) {
			//System.out.println("ok");
			//System.out.println(((PointSelection)selection).getSize());
			acp = ((PointSelection)selection).getControlPointArray();
			ap3 = ((PointSelection)selection).getPointArray();
			p3Center = ((PointSelection)selection).getCenter();
		}
		MainFrame.getInstance().setHelpText("drag to rotate");
	}
	
	public void mousePressed(MouseEvent mouseEvent) {
		if (mouseEvent.getButton() == MouseEvent.BUTTON1) {
			compSource = (Component)mouseEvent.getSource();
			Viewport viewport = (Viewport)mouseEvent.getSource();
			iMouseX = mouseEvent.getX();
			iMouseY = mouseEvent.getY();
			compSource.addMouseMotionListener(this);
			//compoundEdit = new CompoundEdit();
			MainFrame.getInstance().setHelpText("hold SHIFT to snap to 45 steps, CTRL to snap to 15 steps");
		}
	}
	
	public void mouseReleased(MouseEvent mouseEvent) {
		if (mouseEvent.getButton() == MouseEvent.BUTTON1) {
			compSource.removeMouseMotionListener(this);
			compoundEdit.addEdit(new MoveControlPointsEdit(MoveControlPointsEdit.ROTATE,acp));
			compoundEdit.end();
			MainFrame.getInstance().getUndoManager().addEdit(compoundEdit);
			MainFrame.getInstance().getMeshToolBar().reset();
			MainFrame.getInstance().getJPatchScreen().full_update();
		}
	}
	
	public void mouseDragged(MouseEvent mouseEvent) {
		int iDeltaX = mouseEvent.getX() - iMouseX;
		int iDeltaY = mouseEvent.getY() - iMouseY;
		//iMouseX = mouseEvent.getX();
		//iMouseY = mouseEvent.getY();
		float fRotationX = iDeltaX / 2;
		float fRotationY = -iDeltaY / 2;
		if (mouseEvent.isShiftDown()) {
			fRotationX = Math.round(fRotationX / 45) * 45;
			fRotationY = Math.round(fRotationY / 45) * 45;
		} else if (mouseEvent.isControlDown()) {
			fRotationX = Math.round(fRotationX / 15) * 15;
			fRotationY = Math.round(fRotationY / 15) * 15;
		}
		m4InvView.set(((Viewport)(mouseEvent.getSource())).getViewDefinition().getMatrix());
		m4InvView.invert();
		//System.out.println(m4View);
		switch(rotateDialog.getAxis()) {
			case RotateDialog.WINDOW_Z:
				v3Axis.set(0,0,1);
				m4InvView.transform(v3Axis);
				axisAngle.set(v3Axis,fRotationX/180*(float)Math.PI);
				rotate(MainFrame.getInstance().getSelection(),axisAngle);
				break;
			case RotateDialog.WINDOW_XY:
				v3Axis.set(1,0,0);
				m4InvView.transform(v3Axis);
				axisAngle.set(v3Axis,fRotationY/180*(float)Math.PI);
				//rotate(MainFrame.getInstance().getSelection(),axisAngle);
				v3Axis.set(0,1,0);
				m4InvView.transform(v3Axis);
				axisAngle2.set(v3Axis,fRotationX/180*(float)Math.PI);
				rotate(MainFrame.getInstance().getSelection(),axisAngle,axisAngle2);
				break;
			case RotateDialog.X:
				v3Axis.set(1,0,0);
				axisAngle.set(v3Axis,fRotationX/180*(float)Math.PI);
				rotate(MainFrame.getInstance().getSelection(),axisAngle);
				break;
			case RotateDialog.Y:
				v3Axis.set(0,1,0);
				axisAngle.set(v3Axis,fRotationX/180*(float)Math.PI);
				rotate(MainFrame.getInstance().getSelection(),axisAngle);
				break;
			case RotateDialog.Z:
				v3Axis.set(0,0,1);
				axisAngle.set(v3Axis,fRotationX/180*(float)Math.PI);
				rotate(MainFrame.getInstance().getSelection(),axisAngle);
				break;
		}	
		MainFrame.getInstance().getJPatchScreen().single_update(compSource);
	}
	
	public void rotate(Selection selection, AxisAngle4f axisAngle) {
		rotate(selection,axisAngle,null);
	}
	
	public void rotate(Selection selection, AxisAngle4f axisAngle, AxisAngle4f axisAngle2) {
		
		Vector3f cv = new Vector3f();
		switch(rotateDialog.getPivot()) {
			case RotateDialog.ORIGIN:
				cv.set(0,0,0);
				break;
			case RotateDialog.CENTER:
				cv.set(p3Center);
				break;
			case RotateDialog.CURSOR:
				cv.set(MainFrame.getInstance().getJPatchScreen().get3DCursor().getPosition());
				break;
		}
		Point3f point = new Point3f();
		for (int i = 0; i < acp.length; i++) {
			ControlPoint cp = acp[i];
			point.set(ap3[i]);
			//if (cp.isHead()) {
				Matrix4f transform = new Matrix4f();
				transform.setIdentity();
				Matrix4f dummy = new Matrix4f();
				//Vector3f cv2 = new Vector3f(cv);
				transform.set(cv);
				dummy.setIdentity();
				dummy.set(axisAngle);
				transform.mul(dummy);
				if (axisAngle2 != null) {
					dummy.setIdentity();
					dummy.set(axisAngle2);
					transform.mul(dummy);
				}
				cv.scale(-1);
				dummy.set(cv);
				cv.scale(-1);
				transform.mul(dummy);
				transform.transform(point);
				//cp.invalidateTangents();
				MainFrame.getInstance().getConstraints().setControlPointPosition(cp,point);
				//cp.setPosition(point);
			//}
		
		}
	}
}
