package net.sf.jpatch.boundary.selection;

import java.util.*;
import javax.vecmath.*;
import net.sf.jpatch.entity.*;
import net.sf.jpatch.boundary.*;
import net.sf.jpatch.auxilary.*;

public class PointSelection extends Selection {
	private static PointSelection ps = new PointSelection(false);
	private static ControlPoint[] acp = new ControlPoint[0];

	protected Collection colCp;
	protected Point3f p3Pivot;
	protected Matrix3f m3Rotation = new Matrix3f();;
	protected Point3f p3CornerA = new Point3f();
	protected Point3f p3CornerB = new Point3f();
	protected ControlPoint cpHot;
	protected boolean bDirection = false;
	protected boolean bCurve = false;
	
	private PointSelection(boolean b) {
	}
	
	public PointSelection() {
		colCp = new HashSet();
		m3Rotation.setIdentity();
	}
	
	public PointSelection(ControlPoint controlPoint) {
		this();
		setControlPoint(controlPoint);
	}
	
	public static Class getPointSelectionClass() {
		return ps.getClass();
	}
	
	public void addPointSelection(PointSelection ps) {
		colCp.addAll(ps.colCp);
	}
	
	public Collection getSelectedControlPoints() {
		return colCp;
	}
	
	public void xorPointSelection(PointSelection ps) {
		for (Iterator i = ps.colCp.iterator(); i.hasNext(); ) {
			ControlPoint cp = (ControlPoint)i.next();
			if (colCp.contains(cp)) {
				colCp.remove(cp);
			} else {
				colCp.add(cp);
			}
		}
		if (colCp.size() == 0) {
			MainFrame.getInstance().setSelection(null);
		}
	}
	
	public void setControlPoint(ControlPoint controlPoint) {
		colCp.clear();
		colCp.add(controlPoint);
		cpHot = controlPoint;
	}
	
	public boolean isSingle() {
		return (colCp.size() == 1);
	}
	
	public ControlPoint[] getControlPointArray() {
		return (ControlPoint[])colCp.toArray(acp);
	}
	
	public void setHotCp(ControlPoint cp) {
		cpHot = cp;
	}
	
	public ControlPoint getHotCp() {
		return cpHot;
	}
	
	public Point3f[] getPointArray() {
		Point3f[] ap3 = new Point3f[colCp.size()];
		int p = 0;
		for (Iterator it = colCp.iterator(); it.hasNext();) {
			//ControlPoint cp = (ControlPoint)it.next();
			//Point3f p3 = cp.getPosition();
			//ap3[p++] = new Point3f(p3);
			ap3[p++] = new Point3f(((ControlPoint)it.next()).getPosition());
		}
		return ap3;
	}
	
	public ControlPoint getControlPoint() {
		if (isSingle()) {
			return (ControlPoint)colCp.iterator().next();
		} else {
			throw new JPatchException("attempted to get singe ControlPoint but size != 1");
		}
	}
	/*
	public ControlPoint getControlPoint(int index) {
		return (ControlPoint)alCp.get(index);
	}
	*/
	public int getSize() {
		return colCp.size();
	}
	
	public void addControlPoint(ControlPoint controlPoint) {
		colCp.add(controlPoint);
	}
	
	public void removeControlPoint(ControlPoint controlPoint) {
		colCp.remove(controlPoint);
		if (colCp.size() == 0) {
			MainFrame.getInstance().setSelection(null);
		}
	}
	
	public boolean contains(ControlPoint controlPoint) {
		return colCp.contains(controlPoint);
	}
	
	public void applyMaterial(JPatchMaterial material) {
		for (Patch patch = MainFrame.getInstance().getModel().getFirstPatch(); patch != null; patch = patch.getNext()) {
			if (patch.isSelected(this)) {
				patch.setMaterial(material);
			}
		}
	}
	/*
	public void setPivot(Point3f pivot) {
		p3Pivot.set(pivot);
	}
	*/
	public Point3f getPivot() {
		if (p3Pivot == null) {
			p3Pivot = getCenter();
		}
		return p3Pivot;
	}
	
	public Point3f getCornerA() {
		computeBounds();
		return p3CornerA;
	}
	
	public Point3f getCornerB() {
		computeBounds();
		return p3CornerB;
	}
	
	public Matrix3f getRotation() {
		if (m3Rotation == null) {
			m3Rotation = new Matrix3f();
			m3Rotation.setIdentity();
		}
		return m3Rotation;
	}
	
	public void resetPivotToCenter() {
		if (p3Pivot == null) {
			p3Pivot = new Point3f();
		}
		p3Pivot.set(getCenter());
	}
	
	protected void computeBounds() {
		float xMax = -Float.MAX_VALUE;
		float xMin = Float.MAX_VALUE;
		float yMax = -Float.MAX_VALUE;
		float yMin = Float.MAX_VALUE;
		float zMax = -Float.MAX_VALUE;
		float zMin = Float.MAX_VALUE;
		Point3f[] ap3 = getPointArray();
		Point3f p3 = new Point3f();
		Matrix3f m3 = new Matrix3f(m3Rotation);
		m3.invert();
		for (int p = 0; p < ap3.length; p++) {
			p3.set(ap3[p]);
			m3.transform(p3);
			if (p3.x > xMax) xMax = p3.x;
			if (p3.x < xMin) xMin = p3.x;
			if (p3.y > yMax) yMax = p3.y;
			if (p3.y < yMin) yMin = p3.y;
			if (p3.z > zMax) zMax = p3.z;
			if (p3.z < zMin) zMin = p3.z;
		}
		p3CornerA.set(xMin,yMin,zMin);
		p3CornerB.set(xMax,yMax,zMax);
	}
	
	public Point3f getCenter() {
		Point3f p3 = new Point3f();
		computeBounds();
		p3.interpolate(p3CornerA,p3CornerB,0.5f);
		m3Rotation.transform(p3);
		return p3;
	}
	
	/**
	 * returns StringBuffer containing an XML representation of the curve, used to save
	 * models in XML format
	 *
	 * @return A StringBuffer containing an XML representation of this curve
	 */
	public StringBuffer xml(int tabs) {
		StringBuffer sbIndent = XMLutils.indent(tabs);
		StringBuffer sbLineBreak = XMLutils.lineBreak();
		StringBuffer sb = new StringBuffer();
		sb.append(sbIndent).append("<selection name=\"" + getName() + "\">");
		//int size = getType();
		//int p = 0;
		for (Iterator it = colCp.iterator(); it.hasNext();) {
			int i = ((ControlPoint) it.next()).getXmlNumber();
			if (i != -1) {
				sb.append(i);
				if (it.hasNext()) {
					sb.append(",");
				}
			}
		}
		sb.append("</selection>").append(sbLineBreak);
		return sb;
	}
	
	public boolean getDirection() {
		return bDirection;
	}
	
	public boolean isCurve() {
		return bCurve && isSingle();
	}
	
	public void nextPoint() {
		if (!bCurve) {
			bCurve = true;
		} else {
			if (isSingle()) {
				if (bDirection) {
					ControlPoint cp = getControlPoint();
					if (cp.getPrevAttached() != null) {
						setControlPoint(cp.getPrevAttached());
					} else {
						setControlPoint(cp.getHead());
						bCurve = false;
					}
				}
				bDirection = !bDirection;
			}
		}
		//System.out.println(getControlPoint() + " " + getControlPoint().getPrevAttached() + " " + bDirection);
	}
}
