package net.sf.jpatch.control.edit;

import java.util.*;
import javax.vecmath.*;
import net.sf.jpatch.auxilary.*;
import net.sf.jpatch.boundary.*;
import net.sf.jpatch.entity.*;
import net.sf.jpatch.boundary.selection.*;

/**
* Undoable Clone Edit
**/
public class CloneEdit extends JPatchCompoundEdit {

	static private int iNum = 1;
	/**
	* Creates and performes the edit
	* @param  A  ControlPoint A
	* @param  B  ControlPoint B
	*/
	public CloneEdit(ControlPoint[] acp) {

		/*
		 * create set of points to clone
		 */

		//System.out.print("clone " + acp.length);
		Set setCPs = new HashSet();
		for (int i = 0; i < acp.length; i++) {
			ControlPoint[] stack = acp[i].getStack();
			for (int s = 0; s < stack.length; s++) {
				//if (stack[s].isHook()) {
				//	boolean add = true;
				//	loop1:
				//	for (ControlPoint cp = stack[s].getStart(); cp != null; cp = cp.getNext()) {
				//		boolean part = false;
				//		loop2:
				//		for (int p = 0; p < acp.length; p++) {
				//			if (acp[p] == cp) {
				//				part = true;
				//				break loop2;
				//			}
				//		}
				//		if (!part) {
				//			add = false;
				//			System.out.println("*");
				//			break loop1;
				//		}
				//	}
				//	if (add) {
				//		setCPs.add(stack[s]);
				//	}
				//} else { 
				//	setCPs.add(stack[s]);
				//}
				
				if (stack[s].isStartHook()) {
					/* check if hook-curve should be added */
					boolean addHookCurve = false;
					if (JPatchUtils.arrayContains(acp,stack[s].getEnd())) {
						for (ControlPoint cp = stack[s].getStart(); cp.getNext() != null; cp = cp.getNext()) {
							if (JPatchUtils.arrayContains(acp,cp)) {
								addHookCurve = true;
							}
						}
					}
					if (addHookCurve) {
						for (ControlPoint cp = stack[s].getStart(); cp != null; cp = cp.getNext()) {
							setCPs.add(cp);
						}
					}
				} else if (!stack[s].isHook()) {
					setCPs.add(stack[s]);
				}
			}
		}
		//System.out.println(" " + setCPs.size());
		/*
		 * clone controlpoints
		 */

		Model model = MainFrame.getInstance().getModel();
		ArrayList listCPs = new ArrayList();
		ArrayList listNewCurves = new ArrayList();
		HashMap mapCPs = new HashMap();
		PointSelection ps = new PointSelection();

		/* create cloned controlPoints */
		for (Iterator it = setCPs.iterator(); it.hasNext(); ) {
			ControlPoint cpToClone = (ControlPoint) it.next();
			ControlPoint cpClone = new ControlPoint(cpToClone);
			mapCPs.put(cpToClone, cpClone);
		}

		/* connect cloned controlPoints */
		for (Iterator it = setCPs.iterator(); it.hasNext(); ) {
			ControlPoint cpToClone = (ControlPoint) it.next();
			ControlPoint cpClone = (ControlPoint) mapCPs.get(cpToClone);
			if (cpToClone.isHook()) {
				/* if we are a hook, connect to prev and next */
				ControlPoint cp;
				boolean valid;
				ControlPoint cpTargetHook;
				
				cp = cpToClone.getPrev();
				if (cp != null) {
					cpTargetHook  = cp.getPrevAttached();
					valid = (cpTargetHook == null) ? false : JPatchUtils.arrayContains(acp,cpTargetHook.getNext()) || JPatchUtils.arrayContains(acp,cpTargetHook.getPrev());
					while (cp.getPrev() != null && !JPatchUtils.arrayContains(acp,cp) && !valid) {
						cp = cp.getPrev();
					}
				}
				cpClone.setPrev((ControlPoint) mapCPs.get(cp));
				
				cp = cpToClone.getNext();
				if (cp != null) {
					cpTargetHook  = cp.getPrevAttached();
					valid = (cpTargetHook == null) ? false : JPatchUtils.arrayContains(acp,cpTargetHook.getNext()) || JPatchUtils.arrayContains(acp,cpTargetHook.getPrev());
					while (cp.getNext() != null && !JPatchUtils.arrayContains(acp,cp)) {
						cp = cp.getNext();
					}
				}
				cpClone.setNext((ControlPoint) mapCPs.get(cp));
			} else {
				/* we're no hook */
				cpClone.setNext((ControlPoint) mapCPs.get(cpToClone.getNext()));
				cpClone.setPrev((ControlPoint) mapCPs.get(cpToClone.getPrev()));
			}
			cpClone.setNextAttached((ControlPoint) mapCPs.get(cpToClone.getNextAttached()));
			cpClone.setPrevAttached((ControlPoint) mapCPs.get(cpToClone.getPrevAttached()));
			cpClone.setParentHook((ControlPoint) mapCPs.get(cpToClone.getParentHook()));
			cpClone.setChildHook((ControlPoint) mapCPs.get(cpToClone.getChildHook()));
			cpClone.setHookPos(cpToClone.getHookPos());
		}
		
		/* check for loops, add curves to model*/
		for (Iterator it = setCPs.iterator(); it.hasNext(); ) {
			ControlPoint cpToClone = (ControlPoint) it.next();
			ControlPoint cpClone = (ControlPoint) mapCPs.get(cpToClone);
			
			/* check if the cloned curve is closed */
			if (cpToClone.getLoop()) {
				boolean bLoop = false;
				loop:
				for (ControlPoint cp = cpClone.getNext(); cp != null; cp = cp.getNext()) {
					if (cp == cpClone) {
						bLoop = true;
						break loop;
					}
				}
				cpClone.setLoop(bLoop);
			}

			/* add a curve if we are a start point AND have a next point */
			if ((cpClone.getLoop() || cpClone.getPrev() == null) && cpClone.getNext() != null) {
				//model.addCurve(cpClone);
				Curve curve = new Curve(cpClone);
				addEdit(new CreateCurveEdit(curve));	// create a new curve
				addEdit(new ValidateCurveEdit(curve)); // validate new curve
			}
			
			/* detach empty points */
			if (cpClone.getNext() == null && cpClone.getPrev() == null) {
				if (cpClone.getNextAttached() != null) {
					cpClone.getNextAttached().setPrevAttached(cpClone.getPrevAttached());
				}
				if (cpClone.getPrevAttached() != null) {
					cpClone.getPrevAttached().setNextAttached(cpClone.getNextAttached());
				}
			}
		}

		//Point3f box = new Point3f(ps.getCornerB());
		//box.sub(ps.getCornerA());
		//box.scale(0.1f);
		Point3f pos = new Point3f();
		for (Iterator it = setCPs.iterator(); it.hasNext(); ) {
			ControlPoint cpToClone = (ControlPoint) it.next();
			ControlPoint cpClone = (ControlPoint) mapCPs.get(cpToClone);
			if (cpClone.isHead() && (cpClone.getNext() != null || cpClone.getPrev() != null)) {
			//if (cpClone.isHead() && !cpClone.isHook()) {
				/* if we are a head, add us to the new selection */
				ps.addControlPoint(cpClone);
				pos.set(cpToClone.getPosition());
				//pos.x += 5;
				//pos.y -= 5;
				cpClone.setPosition(pos);
				cpClone.setReferencePosition(pos);
			}
		}

		/* clone patches */
		ArrayList newPatches = new ArrayList();
		ArrayList newMaterials = new ArrayList();
		for (Patch p = model.getFirstPatch(); p != null; p = p.getNext()) {
			ControlPoint[] acpOld = p.getControlPoints();
			ControlPoint[] acpNew = new ControlPoint[acpOld.length];
			boolean addPatch = true;
			loop:
			for (int n = 0; n < acpOld.length; n++) {
				if (mapCPs.get(acpOld[n]) == null) {
					addPatch = false;
					break loop;
				} else {
					acpNew[n] = (ControlPoint) mapCPs.get(acpOld[n]);
				}
			}
			if (addPatch) {
				newPatches.add(acpNew);
				newMaterials.add(p.getMaterial());
				if (acpNew.length == 10) {
					ControlPoint[] acp5 = new ControlPoint[] {
						acpNew[0].trueHead(),
						acpNew[2].trueHead(),
						acpNew[4].trueHead(),
						acpNew[6].trueHead(),
						acpNew[8].trueHead()
						};
					model.getCandidateFivePointPatchList().add(acp5);
				}
			}
		}
		Iterator itMat = newMaterials.iterator();
		for (Iterator it = newPatches.iterator(); it.hasNext(); ) {
			Patch patch = new Patch((ControlPoint[]) it.next());
			patch.setMaterial((JPatchMaterial) itMat.next());
			addEdit(new AddPatchEdit(patch));
			//model.addPatch((ControlPoint[]) it.next());
			//model.getLastPatch().setMaterial((JPatchMaterial) itMat.next());
		}

		/* add selection */
		if (ps.getSize() > 0) {
			ps.setName("clone #" + iNum++);
			addEdit(new ChangeSelectionEdit(ps));
			addEdit(new AddSelectionEdit(ps));
		}
	}
}
