package net.sf.jpatch.control.edit;

import javax.swing.undo.*;
import net.sf.jpatch.boundary.*;
import net.sf.jpatch.entity.*;

/**
 *  Thois edit removes a ControlPoint and all appended ControlPoints.
 *  It will create a new CompoundEdit with one RemoveControlPointFromCurveEdit
 *  for each attached ControlPoint
 *
 * @author     aledinsk
 * @created    27. Dezember 2003
 */

public class DeleteHeadEdit extends JPatchCompoundEdit {

	/**
	 * Constructor
	 *
	 * @param  cp ControlPoint to be deleted
	 */
	public DeleteHeadEdit(ControlPoint cp) {
		super("Remove Controlpoint");
		/*
		 * loop through all attached ControlPoints, starting from head
		 */
		ControlPoint[] acpStack = cp.getStack();
		for (int n = 0; n < acpStack.length; n++) {
			cp = acpStack[n];
			Curve curve = acpStack[n].getCurve();	// get the curve we're on
			if (curve != null) {
				/*
				 * if the curve is open
				 */
				if (!curve.isClosed()) {
					/*
					 * if we are not the first or second point on the curve
					 */
					if (cp.getPrev() != null && cp.getPrev().getPrev() != null) {
						addEdit(new DeleteControlPointFromCurveEdit(cp));		// delete the cp
						/*
						 * if we are not the last or second last point on the curve
						 */
						if (cp.getNext() != null && cp.getNext().getNext() != null) {
							addEdit(new CreateCurveEdit(new Curve(cp.getNext())));	// create a new curve
						}
					}

					/*
					 * if we are the first or second point, but not the last or second last one
					 */
					else if (cp.getNext() != null && cp.getNext().getNext() != null) {
						addEdit(new DeleteControlPointFromCurveEdit(cp));		// delete the cp
						addEdit(new ChangeCurveStartEdit(curve,cp.getNext()));		// change curve start
					}
					
					
					//else if (cp.getNext() == null && cp.getPrev() == null) {
					//	addEdit(new DeleteControlPointFromCurveEdit(cp));
					//}
					
					/*
					 * or else, (only two points left)
					 */
					else if (curve.getModel() != null) {
						/*
						 * detach all points on the curve
						 */
						ControlPoint[] acpCurve = curve.getControlPointArray();
						for (int c = 0; c < acpCurve.length; c++) {
							addEdit(new DetachControlPointEdit(acpCurve[c]));
							addEdit(new ChangeCPCurveEdit(acpCurve[c],null));
						}
						/*
						 * and remove the curve from the model
						 */
						addEdit(new RemoveCurveFromModelEdit(curve));
					}
				}

				/*
				 * else (if the curve is closed)
				 */
				else {
					/*
					 * if there are less than 3 points on the curve
					 */
					if (curve.getLength() < 3 && curve.getModel() != null) {
						
						/*
						 * detach all points
						 */
						ControlPoint[] acpCurve = curve.getControlPointArray();
						for (int c = 0; c < acpCurve.length; addEdit(new DetachControlPointEdit(acpCurve[c++]))); 
						
						/*
						 * remove the curve from the model
						 */
						addEdit(new RemoveCurveFromModelEdit(curve));
					} else {
						ControlPoint cpNewStart = cp.getNext();			// the curve will be opened, this is the new start
						addEdit(new DeleteControlPointFromCurveEdit(cp));	// delete our point
						if (cpNewStart != null) {
							addEdit(new ChangeCurveStartEdit(curve, cpNewStart));	// change curve start
							/*
							 * loop thurough curve and clear loop flags
							 */
							//for (ControlPoint cpX = cpNewStart.getNext(); cpX != cpNewStart && cpX != null; cpX = cpX.getNext()) {
							//	if (cpX.getLoop()) {
							//		addEdit(new ChangeControlPointLoopEdit(cpX, false));
							//	}
							//}
							addEdit(new ChangeControlPointLoopEdit(cpNewStart.getEnd(), false));
						} else {
							System.err.println("**********AAAAAAAAAAAAAAAAAAAAAAAAAAA*************");
							for (ControlPoint cpCurve = curve.getStart(); cpCurve != null; cpCurve = cpCurve.getNextCheckNextLoop()) {
								if (cpCurve != cp) {
									addEdit(new DetachControlPointEdit(cpCurve));
								}
							}
							addEdit(new RemoveCurveFromModelEdit(curve));
						}
					}
				}
			}
		}
	}
}
