package net.sf.jpatch.control.edit;

import java.util.*;
import javax.swing.undo.*;
import net.sf.jpatch.boundary.*;
import net.sf.jpatch.entity.*;

public class RemoveCurveSegmentEdit extends JPatchCompoundEdit {
	
	public RemoveCurveSegmentEdit(ControlPoint cp) {
		Curve curve = cp.getCurve();
		ControlPoint cpNext = cp.getNext();
		
		/* remove hooks... */
		if (cp.getChildHook() != null) {
			//System.out.println("*");
			if (cp.getChildHook().getCurve() != null) {
				addEdit(new RemoveCurveEdit(cp.getChildHook().getCurve()));
			}
			addEdit(new ChangeCPChildHookEdit(cp, null));
		}
		
		/* remove the curve segment */
		addEdit(new ChangeControlPointPrevEdit(cpNext,null));
		addEdit(new ChangeControlPointNextEdit(cp,null));
		
		/* remove patches */
		ArrayList patches1 = MainFrame.getInstance().getModel().getPatchesContaining(cp);
		ArrayList patches2 = MainFrame.getInstance().getModel().getPatchesContaining(cpNext);
		for (Iterator it = patches1.iterator(); it.hasNext();) {
			Patch patch = (Patch) it.next();
			if (patches2.contains(patch) && patch.getModel() != null) {
				addEdit(new RemovePatchFromModelEdit(patch));
			}
		}
		
		/* check if curve is closed */
		if (curve.isClosed()) {
			
			/* change curve start */
			addEdit(new ChangeControlPointLoopEdit(curve.getStart(),false));
			addEdit(new ChangeCurveStartEdit(cp.getCurve(),cpNext));
		} else {
			
			/* check if we are the first cp on the curve */
			if (cp.getPrev() == null) {
				
				
				/* check if the next cp is the last on the curve */
				if (cpNext.getNext() == null) {
				
					/* remove hooks */
					if (cpNext.getNextAttached() != null && cpNext.getNextAttached().isHook()) {
						Curve hookCurve = cpNext.getNextAttached().getCurve();
						addEdit(new RemoveControlPointFromSelectionsEdit(cpNext.getNextAttached()));
						addEdit(new RemoveControlPointFromCurveEdit(cpNext.getNextAttached()));
						if (hookCurve.getLength() == 2) {
							addEdit(new ChangeCPChildHookEdit(hookCurve.getStart().getParentHook(),null));
							//addEdit(new RemoveCurveFromModelEdit(hookCurve));
							addEdit(new RemoveCurveEdit(hookCurve));
						}
					}
					
					/* remove the entire curve */
					addEdit(new RemoveCurveFromModelEdit(curve));
					addEdit(new DetachControlPointEdit(cp));
					addEdit(new DetachControlPointEdit(cpNext));
					addEdit(new RemoveControlPointFromSelectionsEdit(cp));
					addEdit(new RemoveControlPointFromSelectionsEdit(cpNext));
				} else {
					
					/* remove the controlpoint */
					addEdit(new RemoveControlPointFromSelectionsEdit(cp));
					addEdit(new DetachControlPointEdit(cp));
					addEdit(new ChangeCurveStartEdit(cp.getCurve(),cpNext));
					addEdit(new ChangeCPCurveEdit(cp,null));
				}
				
				/* remove hooks */
				if (cp.getNextAttached() != null && cp.getNextAttached().isHook()) {
					Curve hookCurve = cp.getNextAttached().getCurve();
					addEdit(new RemoveControlPointFromCurveEdit(cp.getNextAttached()));
					addEdit(new RemoveControlPointFromSelectionsEdit(cp.getNextAttached()));
					if (hookCurve.getLength() == 2) {
						addEdit(new ChangeCPChildHookEdit(hookCurve.getStart().getParentHook(),null));
						addEdit(new RemoveCurveEdit(hookCurve));
					}
				}
			} else {
				
				/* check if the next cp is not the last on the curve */
				if (cpNext.getNext() != null) {
				
					/* add a new curve */
					Curve newCurve = new Curve(cpNext);
					addEdit(new CreateCurveEdit(newCurve));
					addEdit(new ValidateCurveEdit(newCurve));
				} else {
					/* remove hooks */
					if (cpNext.getNextAttached() != null && cpNext.getNextAttached().isHook()) {
						Curve hookCurve = cpNext.getNextAttached().getCurve();
						addEdit(new RemoveControlPointFromSelectionsEdit(cpNext.getNextAttached()));
						addEdit(new RemoveControlPointFromCurveEdit(cpNext.getNextAttached()));
						if (hookCurve.getLength() == 2) {
							addEdit(new ChangeCPChildHookEdit(hookCurve.getStart().getParentHook(),null));
							//addEdit(new RemoveCurveFromModelEdit(hookCurve));
							addEdit(new RemoveCurveEdit(hookCurve));
						}
					}
					
					/* remove the next cp */
					addEdit(new DetachControlPointEdit(cpNext));
					addEdit(new RemoveControlPointFromSelectionsEdit(cpNext));
				}
			}
		}
	}
}
