package net.sf.jpatch.control.edit;

import javax.swing.undo.*;
import net.sf.jpatch.boundary.*;
import net.sf.jpatch.entity.*;

/**
 *  Thois edit removes a ControlPoint and all appended ControlPoints.
 *  It will create a new CompoundEdit with one RemoveControlPointFromCurveEdit
 *  for each attached ControlPoint
 *
 * @author     aledinsk
 * @created    27. Dezember 2003
 */

public class RemoveHeadEdit extends JPatchCompoundEdit {

	/**
	 * Constructor for RemoveControlPointEdit
	 *
	 * @param  cp  ControlPoint to be removed
	 */
	public RemoveHeadEdit(ControlPoint cp) {
		super("RemoveHeadEdit");
		System.out.println("RemoveHeadEdit(" + cp + ")");
		/*
		 * loop through all attached ControlPoints, starting from head
		 */
		ControlPoint[] acpStack = cp.getStack();
		for (int n = 0; n < acpStack.length; n++) {
			Curve curve = acpStack[n].getCurve();
			if (curve != null) {
				
				/*
				 * if the curve's lenght is < 3 remove the whole curve
				 */
				if (curve.getLength() < 3 && curve.getModel() != null) {
					/*
					 * remove the curve from the model
					 */
					addEdit(new RemoveCurveFromModelEdit(curve));
					/*
					 * detach all points on the deleted curve
					 */
					ControlPoint[] acpCurve = curve.getControlPointArray();
					for (int c = 0; c < acpCurve.length; c++) {
						addEdit(new DetachControlPointEdit(acpCurve[c]));
						addEdit(new ChangeCPCurveEdit(acpCurve[c],null));
					}
				}
				
				/*
				 * or else just remove the controlPoint
				 */
				else {
					addEdit(new RemoveControlPointFromCurveEdit(acpStack[n]));
				}
			}
		}
	}
}
