package net.sf.jpatch.entity;

import javax.vecmath.*;
import javax.swing.tree.*;
import java.util.*;
import net.sf.jpatch.auxilary.*;

public class Bone
implements TreeNode {
	private static int iNumber;
	private String strName;
	private Model model;
	private Point3f p3Start = new Point3f();
	private Vector3f v3Extent = new Vector3f();
	private float fStartRadius;
	private float fEndRadius;
	private Bone boneParent;
	private Bone boneNext;
	private Bone bonePrev;
	private int iChilds = 0;
	private Point3f[] ap3Representation = new Point3f[7];
	private boolean bSelected;
	
	public Bone(Model model, Point3f start, Point3f end) {
		this.model = model;
		for (int p = 0; p < 7; p++) {
			ap3Representation[p] = new Point3f();
		}
		p3Start.set(start);
		v3Extent.sub(end,start);
		float radius = v3Extent.length()/3f;
		fStartRadius = radius;
		fEndRadius = radius;
		strName = "bone" + iNumber++;
	}
	
	public Bone(Model model,Bone parent) {
		this(model, parent.getStart(),parent.getStart());
		boneParent = parent;
		boneParent.iChilds++;
	}
	
	public String toString() {
		return strName;
	}
	
	/*
	* TreeNode interface implementation
	*/
	
	public Enumeration children() {
		return new ArrayEnumeration(getChilds());
	}
	
	public boolean getAllowsChildren() {
		return true;
	}
	
	public void setSelected(boolean selected) {
		bSelected = selected;
	}
	
	public boolean isSelected() {
		return bSelected;
	}
	
	public TreeNode getChildAt(int childIndex) {
		Bone[] aboneChildren = getChilds();
		return aboneChildren[childIndex];
	}
	
	public int getChildCount() {
		return getNumberOfChilds();
	}
	
	public int getIndex(TreeNode node) {
		Bone bone = (Bone)node;
		Bone[] aboneChildren = getChilds();
		for (int b = 0; b < aboneChildren.length; b++) {
			if (aboneChildren[b] == bone) {
				return b;
			}
		}
		return -1;
	}
	
	public TreeNode getParent() {
		return boneParent;
	}
	
	public boolean isLeaf() {
		return false;
		//return (iChilds == 0);
	}
	
	/*
	* TreeNode interface implementation end
	*/
	
	public Bone getRoot() {
		if (boneParent == null) {
			return this;
		} else {
			return boneParent.getRoot();
		}
	}
	
	public Point3f getStart() {
		if (boneParent != null) {
			return boneParent.getEnd();
		} else {
			return p3Start;
		}
	}
	
	public Point3f getEnd() {
		Point3f p3End = new Point3f(getStart());
		p3End.add(v3Extent);
		return p3End;
	}
	
	public Bone[] getChilds() {
		if (iChilds > 0) {
			Bone[] aboneChild = new Bone[iChilds];
			int c = 0;
			for (Bone bone = model.getFirstBone(); bone != null; bone = bone.getNext()) {
				if (bone.boneParent == this) {
					aboneChild[c++] = bone;
				}
			}
			return aboneChild;
		}
		return null;
	}
	
	public int getNumberOfChilds() {
		return iChilds;
	}
	
	public void setStart(Point3f start) {
		if (boneParent == null) {
			p3Start = start;
		}
	}
	
	public void setParent(Bone parent) {
		boolean loop = false;
		for (Bone bone = parent; bone != null; bone = bone.boneParent) {
			loop |= (bone == this);
		}
		if (!loop) {	
			if (boneParent != null) {
				boneParent.iChilds--;
			}
			if (parent != null) {
				parent.iChilds++;
			} else {
				p3Start.set(getStart());
			}
			boneParent = parent;
		}
	}
	
	public void setEnd(Point3f end) {
		v3Extent.sub(end,getStart());
	}
	
	public Vector3f getExtent() {
		return v3Extent;
	}
	
	public float getLength() {
		return v3Extent.length();
	}
	
	public Bone getParentBone() {
		return boneParent;
	}
	
	public Bone getNext() {
		return boneNext;
	}
	/*
	public Bone getPrev() {
		return bonePrev;
	}
	*/
	public void insertBefore(Bone firstBone) {
		boneNext = firstBone;
		if (firstBone != null) {
			firstBone.bonePrev = this;
		}
	}
	
	private void destroy(boolean reconnect) {
		if (model.getFirstBone() == this) {
			model.setFirstBone(boneNext);
		}
		if (iChilds > 0) {
			Bone[] aboneChild = getChilds();
			for (int b = 0; b < aboneChild.length; b++) {
				if (reconnect) {
					aboneChild[b].boneParent = boneParent;
					boneParent.iChilds++;
					aboneChild[b].v3Extent.add(v3Extent);
				} else {
					aboneChild[b].boneParent = null;
					aboneChild[b].p3Start.set(getEnd());
				}
			}
		}
		if (boneParent != null) {
			boneParent.iChilds--;
		}
		if (bonePrev != null) {
			bonePrev.boneNext = boneNext;
		}
		if (boneNext != null) {
			boneNext.bonePrev = bonePrev;
		}
	}
	
	public void delete() {
		destroy(false);
	}
	
	public void remove() {
		destroy(true);
	}
	/*	
	public Point3f[] getRepresentation() {
		Point3f p3Helper = new Point3f();
		ap3Representation[0].set(getStart());
		ap3Representation[5].set(getEnd());
		if (v3Extent.x != 0 || v3Extent.y != 0 || v3Extent.z != 0) {
			Vector3f v3Helper = (v3Extent.x == 0) ? new Vector3f(1,0,0) : new Vector3f(0,1,0);
			v3Helper.cross(v3Helper,v3Extent);
			v3Helper.normalize();
			float radius = v3Extent.length() * 0.075f;
			v3Helper.scale(radius);
			p3Helper.interpolate(ap3Representation[0],ap3Representation[5],0.1f);
			AxisAngle4f axisAngle = new AxisAngle4f(v3Extent,(float)Math.PI/2f);
			Matrix3f m3Rotation = new Matrix3f();
			m3Rotation.set(axisAngle);
			for (int p = 1; p < 5; p++) {
				ap3Representation[p].set(p3Helper);
				ap3Representation[p].add(v3Helper);
				m3Rotation.transform(v3Helper);
			}
		} else {
			for (int p = 1; p < 5; p++) {
				ap3Representation[p].set(ap3Representation[0]);
			}
		}
		ap3Representation[6].x = (bSelected) ? 1 : 0;
		return ap3Representation;
	}
	*/
}
