package net.sf.jpatch.render;

import java.util.*;
import javax.vecmath.*;
import net.sf.jpatch.entity.*;

public class RPhongColorShader implements RShader {
	protected Point3d point = new Point3d();
	protected Vector3d normal = new Vector3d();
	protected MaterialProperties mp;
	
	public Color3f shade(double[] shaderParams, MaterialProperties mp, ArrayList lights) {
		
		point.set(shaderParams[RConstants.POSITION_X],shaderParams[RConstants.POSITION_Y],1d / shaderParams[RConstants.POSITION_R]);
		normal.set(shaderParams[RConstants.NORMAL_X],shaderParams[RConstants.NORMAL_Y],shaderParams[RConstants.NORMAL_Z]);
		normal.normalize();
		this.mp = mp;
		
		Vector3d D = new Vector3d(point);
		D.normalize();
		if (D.dot(normal) >= 0) {
			normal.x = -normal.x;
			normal.y = -normal.y;
			normal.z = -normal.z;
		}
		
		return shade(lights);
	}
	
	private Color3f shade(ArrayList lights) {
		Color3f color = new Color3f();
		for (Iterator it = lights.iterator(); it.hasNext(); ) {
			RLight light = (RLight) it.next();
			
			boolean doAmbient = (mp.ambient != 0);
			boolean doDiffuse = (mp.diffuse != 0);
			boolean doSpecular = (light.highlight() && mp.specular != 0);

			Color3f intensity = null;
			Vector3d direction = null;
			
			if (doAmbient || doDiffuse || doSpecular) {
				intensity = light.lightIntensity(point);
				direction = light.lightDirection(point);
			}
				
			if (direction == null) {
				/* ambient light! */
				if (doAmbient) {
					color.add(computeAmbient(intensity));
				}
			} else {
				/* some kind of directed light */
				if (doDiffuse) {
					color.add(computeDiffuse(intensity,direction));
				}
				if (doSpecular) {
					color.add(computeSpecular(intensity,direction));
				}
			}
		}
		return color;
	}
	
	protected Color3f computeAmbient(Color3f intensity) {
		Color3f ambient = new Color3f(intensity);
		ambient.scale(mp.ambient);
		ambient.x *= mp.red;
		ambient.y *= mp.green;
		ambient.z *= mp.blue;
		return ambient;
	}
	
	protected Color3f computeDiffuse(Color3f intensity, Vector3d direction) {
		Color3f diffuse = new Color3f(1,1,1);
		diffuse.scale((float) direction.dot(normal));
		diffuse.clamp(0,1);
		diffuse.x *= intensity.x * mp.red;
		diffuse.y *= intensity.y * mp.green;
		diffuse.z *= intensity.z * mp.blue;
		return diffuse;
	}
	
	protected Color3f computeSpecular(Color3f intensity, Vector3d direction) {
		Vector3d LV = new Vector3d();
		Vector3d P = new Vector3d(point);
		P.normalize();
		LV.sub(direction,P);
		LV.scale(1d/Math.sqrt(LV.dot(LV)));
		LV.normalize();
		double spec = LV.dot(normal);
		Color3f specular = new Color3f(intensity);
		specular.scale(Math.max((float)Math.pow(spec,1f/mp.roughness) * mp.specular,0));
		specular.clamp(0,1000f);
		return specular;
	}
}
