package net.sf.jpatch.render;

import java.awt.*;
import java.awt.geom.*;
import java.awt.image.*;
import java.io.*;
import java.util.*;
import javax.swing.*;
import javax.vecmath.*;
import net.sf.jpatch.entity.*;

public class RRenderer extends JPanel {
	private ArrayList listPatches = new ArrayList();
	private int iWidth;
	private int iHeight;
	private int iSupersample;
	private int iFullWidth;
	private int iFullHeight;
	private Color3f c3Background;
	private RCamera camera = new RCamera();
	private BufferedImage image;
	private int[] aiFrameBuffer;
	private RShader shader;
	private ArrayList listLights = new ArrayList();
	
	public RRenderer() {
		setup();
	}
	
	public void setup() {
		iWidth = 640;
		iHeight = 480;
		iSupersample = 1;
		camera.setPosition(0,0,-50);
		camera.pointAt(0,1,-30);
		camera.setFocalLength(50);
		c3Background = new Color3f(0,0,1);
		shader = new RPhongColorShader();
		//listLights.add(new RAmbientLight(new Color3f(2.5f,2.5f,2.5f)));
		listLights.add(new RDistantLight(new Color3f(1,0,0), new Vector3f(1,-1,1), true));
		init();
	}
	
	public void init() {
		iFullWidth = iWidth * iSupersample;
		iFullHeight = iHeight * iSupersample;
		float f = camera.getFocalLength(iFullWidth);
		RBezierPatch.init(iFullWidth,iFullHeight,5,f);
		RTriangle.init(iFullWidth,iFullHeight);
		image = new BufferedImage(iWidth, iHeight, BufferedImage.TYPE_INT_RGB);
		aiFrameBuffer = ((DataBufferInt) image.getRaster().getDataBuffer()).getData();
		setPreferredSize(new Dimension(iWidth,iHeight));
	}
	
	public void paint(Graphics g) {
		g.drawImage(image,0,0,null);
	}
	public void loadPatchesFile(String filename) {
		BufferedReader brFile;
		String strLine;
		String[] astrPart;
		double x,y,z;
		MaterialProperties mp = new MaterialProperties();
		Matrix4f m4View = camera.getMatrix();
		System.out.println(m4View);
		try {
			brFile = new BufferedReader(new FileReader(filename));
			while ((strLine = brFile.readLine()) != null) {
				
				if (strLine.startsWith("m")) {
					
					/* read material properties */
					mp = new MaterialProperties();
					astrPart = strLine.split("\\s");
					mp.red = (new Float(astrPart[1])).floatValue();
					mp.green = (new Float(astrPart[2])).floatValue();
					mp.blue = (new Float(astrPart[3])).floatValue();
					mp.ambient = (new Float(astrPart[4])).floatValue();
					mp.diffuse = (new Float(astrPart[5])).floatValue();
					mp.brilliance = (new Float(astrPart[6])).floatValue();
					mp.specular = (new Float(astrPart[7])).floatValue();
					mp.roughness = (new Float(astrPart[8])).floatValue();
				} else if (strLine.startsWith("b")) {
				
					/* read patch */
					astrPart = strLine.split("\\s");
					Point3f[] bezierControlPoints = new Point3f[16];
					int n = 1;
					for (int i = 0; i < 16; i++) {
						bezierControlPoints[i] = new Point3f((new Float(astrPart[n++])).floatValue(),(new Float(astrPart[n++])).floatValue(),(new Float(astrPart[n++])).floatValue());
						m4View.transform(bezierControlPoints[i]);
					}
					listPatches.add(new RBezierPatch(bezierControlPoints,mp));
				}
			}
		} catch (IOException e) {
			System.err.println(e);
		}
	}
	
	public void render() {
		
		System.out.println(listPatches.size() + " patches");
		
		/*
		* subdivide
		*/
		float maxSize = 16 * iSupersample;
		
		System.out.println("subdividing...");
		
		int level;
		
		//subdivide:
		//for (level = 0; level < 5; level++) {
		//	ArrayList newPatchList = new ArrayList();
		//	boolean cont = false;
		//	for (Iterator it = listPatches.iterator(); it.hasNext(); ) {
		//		RBezierPatch patch = (RBezierPatch) it.next();
		//		if (!patch.isBehindClippingPlane()) {
		//			/* patch is not behind clipping plane */
		//			if (patch.needsZSplit()) {
		//				/* patch intersects clipping plane -> split */
		//				cont = true;
		//				newPatchList.addAll(patch.split());
		//			} else {
		//				/* patch is infront of clipping plane -> project to screen */
		//				patch.project();
		//				if (patch.isVisible()) {
		//					/* patch is visible */
		//					if (patch.needsSplit() || patch.needsSplit(maxSize)) {
		//						/* patch is partly offscreen or is too large -> slpit */
		//						cont = true;
		//						newPatchList.addAll(patch.split());
		//					} else {
		//						/* patch is fully on screen and small enough -> keep */
		//						newPatchList.add(patch);
		//					}
		//				}
		//			}
		//		}
		//	}
		//	if (cont) {
		//		listPatches = newPatchList;
		//	} else {
		//		break subdivide;
		//	}
		//}
		
		//System.out.println("subdivided " + level + " times");
		/*
		* prepare patches
		*/
		
		System.out.println("preparing...");
		
		for (Iterator it = listPatches.iterator(); it.hasNext(); ) {
			((RBezierPatch) it.next()).prepare();
		}
		
		System.out.println(listPatches.size() + " patches");
		
		/*
		* render scanlines
		*/
		
		System.out.println("rendering...");
		
		
		Graphics g = image.getGraphics();
		for (Iterator it = listPatches.iterator(); it.hasNext(); ) {
			RBezierPatch patch = (RBezierPatch) it.next();
			Rectangle2D.Float rect = patch.getBoundingRectangle();
			rect.x /= iSupersample;
			rect.y /= iSupersample;
			rect.width /= iSupersample;
			rect.height /= iSupersample;
			rect.x += iWidth / 2;
			rect.y = iHeight / 2 - rect.y - rect.height;
			g.drawRect((int) rect.x,(int) rect.y,(int) rect.width,(int) rect.height);
			//Point2D.Float[] p = patch.getProjectedPoints();
			//for (int i = 0; i < 16; i++) {
			//	g.drawRect((int) p[i].x + iWidth / 2,(int) (iHeight / 2 - p[i].y),0,0);
			//}
		}
		repaint();
		
		//for (Iterator it = listPatches.iterator(); it.hasNext(); ) {
		//	RBezierPatch patch = (RBezierPatch) it.next();
		//	patch.tesselate(2);
		//	ArrayList triangles = patch.getTriangleList();
		//	for (Iterator tri = triangles.iterator(); tri.hasNext(); ) {
		//		((RTriangle) tri.next()).drawPreview(g, iSupersample);
		//	}
		//	//repaint();
		//}
		//repaint();
		
		int sl = 0;
		for (int y = 0; y < iHeight; y++) {
			System.out.println(y);
			RScanline[] scanline = new RScanline[iSupersample];
			for (int s = 0; s < iSupersample; s++) {
				scanline[s] = new RScanline(iFullWidth,c3Background);
				ArrayList patchesToRemove = new ArrayList();
				for (Iterator it = listPatches.iterator(); it.hasNext(); ) {
					RBezierPatch patch = (RBezierPatch) it.next();
					//System.out.println("patch bounds: " + patch.getUpperBounds() + " " + patch.getLowerBounds());
					if (sl > patch.getLowerBounds()) {
						patchesToRemove.add(patch);
					} else if (sl >= patch.getUpperBounds()) {
						if (!patch.isReady()) {
							//float f = patch.getSizeOnScreen();
							//int subdiv = (int) (f / iSupersample / 1) + 2;
							//if (subdiv > 4) subdiv = 4;
							patch.tesselate(4);
							
							//ArrayList triangles = patch.getTriangleList();
							//for (Iterator tri = triangles.iterator(); tri.hasNext(); ) {
							//	((RTriangle) tri.next()).drawPreview(g, iSupersample);
							//}
							//repaint();
		
						}
						patch.fillScanline(scanline[s],sl);
					}
				}
				scanline[s].shadeScanline(shader, listLights);
				listPatches.removeAll(patchesToRemove);
				sl++;
			}
			copyScanlinesToBuffer(scanline,y);
			repaint();
		}
		
	}
	
	public void copyScanlinesToBuffer(RScanline[] scanline, int y) {
		int index = y * iWidth;
		Color3f c = new Color3f();
		for (int x = 0; x < iWidth; x++) {
			c.set(0,0,0);
			for (int xx = 0; xx < iSupersample; xx++) {
				for (int yy = 0; yy < iSupersample; yy++) {
					//System.out.println(yy + " " + scanline[yy]);
					c.add(scanline[yy].ac3Color[x * iSupersample + xx]);
				}
			}
			c.scale(1f / (float) (iSupersample * iSupersample));
			c.clamp(0,1);
			float red = c.x * 255;
			float green = c.y * 255;
			float blue = c.z * 255;
			aiFrameBuffer[index++] = 0xFF000000 | (((int)red) << 16) | (((int)green) << 8) | ((int)blue);
		}
	}
}
