package net.sf.jpatch.render;

import java.awt.*;
import javax.vecmath.*;

import net.sf.jpatch.entity.*;

public class RTriangle {
	private static int iWidth;
	private static int iHeight;
	private static int iCenterX;
	private static int iCenterY;
	
	private static double dFocalLength;
	
	private double[] adValueA;
	private double[] adValueB;
	private double[] adValueC;
	
	private double dScreenAx, dScreenAy, dScreenBx, dScreenBy, dScreenCx;
	
	private boolean bUpsideDown;
	
	public static void init(int width, int height) {
		iWidth = width;
		iHeight = height;
		iCenterX = width / 2;
		iCenterY = height / 2;
	}
	
	public RTriangle(double ax, double ay, double bx, double by, double cx, double cy, Point3d pa, Point3d pb, Point3d pc, Vector3d na, Vector3d nb, Vector3d nc) {
	
		/*
		* compute screen coordinates
		*/
		dScreenAx = ax;
		dScreenAy = ay;
		dScreenBx = bx;
		dScreenBy = by;
		dScreenCx = cx;
		
		double ra = 1d / pa.z;
		double rb = 1d / pb.z;
		double rc = 1d / pc.z;
		
		//System.out.println(ra + " " + rb + " " + rc);
		adValueA = new double[RConstants.NUM_VALUES];
		adValueB = new double[RConstants.NUM_VALUES];
		adValueC = new double[RConstants.NUM_VALUES];
		adValueA[RConstants.POSITION_R] = ra;
		adValueA[RConstants.POSITION_X] = pa.x * ra;
		adValueA[RConstants.POSITION_Y] = pa.y * ra;
		adValueB[RConstants.POSITION_R] = rb;
		adValueB[RConstants.POSITION_X] = pb.x * rb;
		adValueB[RConstants.POSITION_Y] = pb.y * rb;
		adValueC[RConstants.POSITION_R] = rc;
		adValueC[RConstants.POSITION_X] = pc.x * rc;
		adValueC[RConstants.POSITION_Y] = pc.y * rc;
		adValueA[RConstants.NORMAL_X] = na.x * ra;
		adValueA[RConstants.NORMAL_Y] = na.y * ra;
		adValueA[RConstants.NORMAL_Z] = na.z * ra;
		adValueB[RConstants.NORMAL_X] = nb.x * rb;
		adValueB[RConstants.NORMAL_Y] = nb.y * rb;
		adValueB[RConstants.NORMAL_Z] = nb.z * rb;
		adValueC[RConstants.NORMAL_X] = nc.x * rc;
		adValueC[RConstants.NORMAL_Y] = nc.y * rc;
		adValueC[RConstants.NORMAL_Z] = nc.z * rc;
				
		bUpsideDown = (dScreenBy < dScreenAy);
		
		//for (int i = 0; i < NUM_VALUES; i++) {
		//	System.out.println(i + " A: " + adValueA[i] + " B: " + adValueB[i] + " C: " + adValueC[i]);
		//}
	}
	
	//public boolean isVisible() {
	//	return (dScreenAx >= 0 && dScreenAx < iWidth && dScreenBx >= 0 && dScreenBx < iWidth && dScreenCx >= 0 && dScreenCx < iWidth && dScreenAy >= 0 && dScreenAy < iHeight && dScreenBy >= 0 && dScreenBy < iHeight);
	//}
	
	public double getLowerBounds() {
		return dScreenBy;
	}
	
	public void drawPreview(Graphics g, int supersample) {
		g.drawLine((int) dScreenAx / supersample,(int) dScreenAy / supersample,(int) dScreenBx / supersample,(int) dScreenBy / supersample);
		g.drawLine((int) dScreenAx / supersample,(int) dScreenAy / supersample,(int) dScreenCx / supersample,(int) dScreenBy / supersample);
		g.drawLine((int) dScreenBx / supersample,(int) dScreenBy / supersample,(int) dScreenCx / supersample,(int) dScreenBy / supersample);
	}
	
	public void fillScanline(RScanline scanline, int y, MaterialProperties mp) {
		//System.out.println("triangle");
		//System.out.println("fillScanline() " + dScreenAy + " " + dScreenBy + " " + y + " " + bUpsideDown);
		/*
		* check if line is visible
		*/
		if ((!bUpsideDown && y >= dScreenAy && y <= dScreenBy) || (bUpsideDown && y <= dScreenAy && y >= dScreenBy)) {
			//System.out.println("fillScanline() +");
			double dy = dScreenBy - dScreenAy;
			double dxleft = dScreenBx - dScreenAx;
			double dxright = dScreenCx - dScreenAx;
			double yy = (double) y - dScreenAy;
			double factor = yy / dy;
			double factor1 = 1 - factor;
			double screenXleft = dScreenAx + dxleft * factor;
			double screenXright = dScreenAx + dxright * factor;
			
			//System.out.println(screenXleft + " " + screenXright);
			/*
			* check if line is visible
			*/
			if (screenXleft < iWidth && screenXright >= 0 && screenXleft < screenXright) {
				//System.out.println("fillScanline() ++");
				double[] left = new double[RConstants.NUM_VALUES];
				double[] right = new double[RConstants.NUM_VALUES];
				
				/*
				* interpolate values
				*/
				for (int i = 0; i < RConstants.NUM_VALUES; i++) {
					double a = adValueA[i] * factor1;
					left[i] = a + adValueB[i] * factor;
					right[i] = a + adValueC[i] * factor;
				}
				
				/*
				* clip to left side screen side or round to exact pixel location
				*/
				if (screenXleft < 0) {
					factor = -screenXleft / (screenXright - screenXleft);
					screenXleft = 0;
				} else {
					double x = Math.ceil(screenXleft);
					factor = (x - screenXleft) / (screenXright - screenXleft);
					screenXleft = x;
				}
				factor1 = 1 - factor;
				for (int i = 0; i < RConstants.NUM_VALUES; i++) {
					left[i] = left[i] * factor1 + right[i] * factor;
				}	
					
				/*
				* clip to right screen side
				*/
				if (screenXright >= iWidth) {
					factor =  (screenXright - iWidth) / (screenXright - screenXleft);
					screenXright = (double) iWidth - 1;
				} else {
					double x = Math.floor(screenXright);
					factor = (screenXright - x) / (screenXright - screenXleft);
					screenXright = x;
				}
				factor1 = 1 - factor;
				for (int i = 0; i < RConstants.NUM_VALUES; i++) {
					right[i] = right[i] * factor1 + left[i] * factor;
				}
				
				/*
				* set up differentials
				*/
				
				double dx = screenXright - screenXleft;
				for (int i = 0; i < RConstants.NUM_VALUES; i++) {
					right[i] = (right[i] - left[i]) / dx;
				}
				
				/*
				* fill scanline
				*/
				int max = (int) screenXright;
				for (int x = (int) screenXleft; x <= max; x++) {
					//System.out.println(left[POSITION_R]);
					if (left[RConstants.POSITION_R] > scanline.aadValue[x][RConstants.POSITION_R]) {
						//System.out.println("zBuffer write");
						double r = left[RConstants.POSITION_R];
						scanline.aadValue[x][RConstants.POSITION_R] = r;
						left[RConstants.POSITION_R] += right[RConstants.POSITION_R];
						for (int i = 1; i < RConstants.NUM_VALUES; i++) {
							scanline.aadValue[x][i] = left[i] / r;
							left[i] += right[i];
						}
						scanline.aMaterialProperties[x] = mp;
						//System.out.println("scanline write");
					}
				}
				//System.out.println(iWidth);
				//if ((int) screenXleft >= 0 && (int) screenXright < iWidth && screenXleft < screenXright) {
				//	//System.out.println(screenXleft + " " + screenXright);
				//	scanline.aMaterialProperties[(int) screenXleft] = materialProperties;
				//	scanline.aMaterialProperties[(int) screenXright] = materialProperties;
				//}
			}
		}
	}
}
