package net.sf.jpatch.renderer;

import java.io.*;
import java.util.*;
import javax.vecmath.*;
import net.sf.jpatch.entity.*;
import net.sf.jpatch.boundary.*;

public class PovrayRenderer extends AbstractRenderer {
	public PovrayRenderer() {
		strRenderer = "Povray";
		strExtension = "inc";
		//strDefaultShaderString = "pigment { color rgb <$red,$green,$blue> }\nfinish { ambient $ambient diffuse $diffuse specular $specular roughness $roughness }\n";
		//strShaderString = "toon(rgb <$red,$green,$blue>)\n";
	}
	
	public void writeToFile(String filename) {
		BufferedWriter file;
		Model model = MainFrame.getInstance().getModel();
		try {
			file = new BufferedWriter(new FileWriter(filename));
			file.write("/*\n");
			file.write("*	JPatch POV-Ray include file\n");
			file.write("*/\n");
			file.write("\n");
			
			for (Iterator iterator = model.getMaterialList().iterator(); iterator.hasNext();) {
				JPatchMaterial material = (JPatchMaterial)iterator.next();
				//System.out.println(material + " " + material.hashCode());
				boolean active = false;
				for(Patch patch = model.getFirstPatch();patch != null; patch = patch.getNext()) {
					//System.out.println(material.hashCode() + " " + patch.getMaterial().hashCode());
					if (patch.getMaterial() == material) {
						//System.out.println("m");
						if (!active) {
							file.write("/*\n");
							file.write("*  " + material.getName() + "\n");
							file.write("*/\n");
							file.write("\n");
							file.write("union {\n");
							active = true;
						}
						Point3f[][] bicubicPatches = patch.bicubicPatches();
						for (int p = 0; p < bicubicPatches.length; p++) {
							Point3f[] bezierControlPoints = bicubicPatches[p];
							file.write("	bicubic_patch { type 1 flatness 0 u_steps 3 v_steps 3 ");
							String strCPs = "";
							for (int c = 0; c < 16; c++) {
								strCPs += ("<" + bezierControlPoints[c].x + "," + bezierControlPoints[c].y + "," + bezierControlPoints[c].z + ">");
								if (c != 15) {
									strCPs += (",");
								}
							}
							file.write(strCPs + " }\n");
							//file.write(" }\n");
						}
					}
				}
				if (active) {
					strShaderString = material.getRenderString("povray","");
					file.write("" + shader(material.getMaterialProperties()));
					file.write("}\n");
				}
				
			}
			
			file.close();
		} catch (IOException e) {
			System.err.println(e.getMessage());
		}
	}		
}
