package net.sf.jpatch.renderer;

import java.io.*;
import java.util.*;
import javax.vecmath.*;
import net.sf.jpatch.entity.*;
import net.sf.jpatch.boundary.*;

public class RibRenderer extends AbstractRenderer {
	public RibRenderer() {
		strRenderer = "RenderMan";
		strExtension = "rib";
		//strDefaultShaderString = "Color [$red $green $blue]\nSurface \"plastic\" \"Ka\" [$ambient] \"Kd\" [$diffuse] \"Ks\" [$specular] \"roughness\" [$roughness]\n";
		//strShaderString = strDefaultShaderString;
	}
	
	public void writeToFile(String filename) {
		BufferedWriter file;
		Model model = MainFrame.getInstance().getModel();
		try {
			file = new BufferedWriter(new FileWriter(filename));
			file.write("##RenderMan RIB-Structure 1.1\n");
			file.write("##Creator JPatch 0.3 beta 1 - http://www.jpatch.com\n");
			file.write("\n\n");
			
			file.write("TransformBegin\n");
			file.write("Basis \"bezier\" 3 \"bezier\" 3\n");
							
			for (Iterator iterator = model.getMaterialList().iterator(); iterator.hasNext();) {
				JPatchMaterial material = (JPatchMaterial)iterator.next();
				boolean active = false;
				for(Patch patch = model.getFirstPatch();patch != null; patch = patch.getNext()) {
					if (patch.getMaterial() == material) {
						if (!active) {
							file.write("#\n");
							file.write("#  " + material.getName() + "\n");
							file.write("#\n");
							file.write("\n");
							file.write("AttributeBegin\n");
							strShaderString = patch.getMaterial().getRenderString("renderman","");
							file.write(shader(patch.getMaterial().getMaterialProperties()));
							active = true;
						}
						Point3f[][] bicubicPatches = patch.bicubicPatches();
						for (int p = 0; p < bicubicPatches.length; p++) {
							Point3f[] bezierControlPoints = bicubicPatches[p];
							
							file.write("Patch \"bicubic\" \"P\" [");
							String strCPs = "";
							for (int c = 0; c < 16; c++) {
								strCPs += (bezierControlPoints[c].x + " " + bezierControlPoints[c].y + " " + bezierControlPoints[c].z);
								if (c != 15) {
									strCPs += (" ");
								}
							}
							file.write(strCPs + "]\n");
						}
					}
				}
				if (active) {
					file.write("AttributeEnd\n");
				}
			}
			file.write("TransformEnd\n");
			file.close();
		} catch (IOException e) {
			System.err.println(e.getMessage());
		}
	}		
}
