package net.sf.jpatch.renderer;

import java.io.*;
import java.util.*;
import javax.vecmath.*;
import net.sf.jpatch.entity.*;
import net.sf.jpatch.boundary.*;
import net.sf.jpatch.auxilary.*;

public class TestRenderer extends AbstractRenderer {
	int subdiv;
	
	public TestRenderer(int subdiv) {
		this.subdiv = subdiv;
	}
	
	public void writeToFile(String filename) {
		BufferedWriter file;
		Model model = MainFrame.getInstance().getModel();
		try {
			file = new BufferedWriter(new FileWriter(filename));
			
			/*
			 * output all coordinates
			 */
			for(Patch patch = model.getFirstPatch();patch != null; patch = patch.getNext()) {
				Point3f[][] bicubicPatches = patch.bicubicPatches();
				for (int p = 0; p < bicubicPatches.length; p++) {
					Point3f[] bezierControlPoints = bicubicPatches[p];
					Tesselator.tesselate(bezierControlPoints,subdiv);
					Point3d[][] p3P = Tesselator.getPointArray();
					Vector3d[][] v3N = Tesselator.getNormalArray();
					for (int u = 0; u < subdiv; u++) {
						for ( int v = 0; v < subdiv; v++) {
							file.write("p " + p3P[u][v].x + " " + p3P[u][v].y + " " + p3P[u][v].z + " ");
							file.write(v3N[u][v].x + " " + v3N[u][v].y + " " + v3N[u][v].z);
							file.newLine();
						}
					}
				}
			}
			
			/*
			 * output all triangles, sorted by material
			 */
			for (Iterator iterator = model.getMaterialList().iterator(); iterator.hasNext();) {
				JPatchMaterial material = (JPatchMaterial)iterator.next();
				MaterialProperties mp = material.getMaterialProperties();
				boolean active = false;
				int p = 0;
				int s = subdiv;
				int s2 = s*s;
				for(Patch patch = model.getFirstPatch();patch != null; patch = patch.getNext()) {
					if (patch.getMaterial() == material) {
						if (!active) {
							/* write material properties */
							file.write("m " + mp.red + " " + mp.green + " " + mp.blue + " ");
							file.write(mp.ambient + " " + mp.diffuse + " " + mp.brilliance + " ");
							file.write(mp.specular + " " + mp.roughness + " ");
							file.newLine();
							active = true;
						}
						for (int i = 0; i < patch.getSize(); i++) {
							int pp = p * s2;
							for (int u = 0; u < s - 1; u++) {
								for (int v = 0; v < s - 1; v++) {
									file.write("f " + (pp + u + v * s) + " " + (pp + u + 1 + v * s) + " " + (pp + u + (v + 1) * s));
									file.newLine();
									file.write("f " + (pp + u + 1 + (v + 1) * s) + " " + (pp + u + (v + 1) * s) + " " + (pp + u + 1 + v * s));
									file.newLine();
								}
							}
							p++;
						}
					} else {
						p += patch.getSize();
					}
				}
			}
			
			file.close();
		} catch (IOException e) {
			System.err.println(e.getMessage());
		}
	}		
}
