package net.sf.jpatch.test;

import javax.vecmath.*;
import net.sf.jpatch.entity.*;
import net.sf.jpatch.boundary.*;
import net.sf.jpatch.control.edit.*;

public class CylinderTest extends Tester {
	private static final int SEGMENTS = 8;
	private static final int SIZE = 5;
	private static final float RADIUS = 10;
	private static final float HEIGHT = 5;
	
	private ControlPoint[][] aacp = new ControlPoint[SEGMENTS][SIZE];
	private Point3f[][] aap3Pos = new Point3f[SEGMENTS][SIZE];
	private boolean[][] aabPoint = new boolean[SEGMENTS][SIZE];
	private boolean[][] aabConnection = new boolean[SEGMENTS][SIZE - 1];
	
	private JPatchCompoundEdit compoundEdit = new JPatchCompoundEdit();
	
	public CylinderTest() {
		super();
		System.out.println("CYLINDER TEST");
		for (int y = 0; y < SIZE; y++) {
			for (int x = 0; x < SEGMENTS; x++) {
				float px = (float) Math.cos(Math.PI * 2f * x / SEGMENTS) * RADIUS;
				float pz = (float) Math.sin(Math.PI * 2f * x / SEGMENTS) * RADIUS;
				float py = HEIGHT * ((float) y - (float) (SIZE - 1) / 2f);
				System.out.println(px + " " + py + " " + pz);
				aap3Pos[x][y] = new Point3f(px,py,pz);
				aabPoint[x][y] = false;
				if (y < SIZE - 1) {
					aabConnection[x][y] = false;
				}
			}
		}
		for (int i = 0; i < 1000; i++) {
			addRandomPoint();
		}
	}
	
	private int upper(int y) {
		return (y > 0) ? y - 1 : -1;
	}
	
	private int lower(int y) {
		return (y < SIZE - 1) ? y + 1 : -1;
	}
	
	private int left(int x) {
		return (x - 1 + SEGMENTS) % SEGMENTS;
	}
	
	private int right(int x) {
		return (x + 1) % SEGMENTS;
	}
	
	private void addRandomPoint() {
		int x = (int) (Math.random() * SEGMENTS);
		int y = (int) (Math.random() * SIZE);
		System.out.println(x + " " + y + aap3Pos[x][y]);
		if (!aabPoint[x][y]) {
			System.out.println("x");
			ControlPoint cpA = new ControlPoint(aap3Pos[x][y]);
			aacp[x][y] = cpA;
			ControlPoint cpB = new ControlPoint();
			cpB.setPrev(cpA);
			cpA.setNext(cpB);
			Curve curve = new Curve(cpA);
			curve.validate();
			compoundEdit.addEdit(new CreateCurveEdit(curve));
			aabPoint[x][y] = true;
			if (upper(y) != -1 && aabPoint[x][upper(y)]) {
				compoundEdit.addEdit(new WeldControlPointsEdit(cpB,aacp[x][upper(y)]));
			} else if (lower(y) != -1 && aabPoint[x][lower(y)]) {
				compoundEdit.addEdit(new WeldControlPointsEdit(cpB,aacp[x][lower(y)]));
			//} else if (left(x) != -1 && aabPoint[left(x)][y]) {
			//	compoundEdit.addEdit(new WeldControlPointsEdit(cpB,aacp[left(x)][y]));
			//} else if (right(x) != -1 && aabPoint[right(x)][y]) {
			//	compoundEdit.addEdit(new WeldControlPointsEdit(cpB,aacp[right(x)][y]));
			} else {
				if (upper(y) != -1) {
					cpB.setPosition(aap3Pos[x][upper(y)]);
					aabPoint[x][upper(y)] = true;
					aacp[x][upper(y)] = cpB;
				} else if (lower(y) != -1) {
					cpB.setPosition(aap3Pos[x][lower(y)]);
					aabPoint[x][lower(y)] = true;
					aacp[x][lower(y)] = cpB;
				//} else if (left(x) != -1) {
				//	cpB.setPosition(aap3Pos[left(x)][y]);
				//	aabPoint[left(x)][y] = true;
				//	aacp[left(x)][y] = cpB;
				//} else if (right(x) != -1) {
				//	cpB.setPosition(aap3Pos[right(x)][y]);
				//	aabPoint[right(x)][y] = true;
				//	aacp[right(x)][y] = cpB;
				}
			}
			undoManager.addEdit(compoundEdit);
			compoundEdit = new JPatchCompoundEdit();
			screen.full_update();
		}
	}
}
